# type: ignore
#!/usr/bin/env python
from __future__ import absolute_import
from __future__ import print_function
from six.moves import range

import sys
import os
import importlib

if '__main__' == __name__:
    dir_path = os.path.join(os.path.dirname(__file__))
    sys.path.insert(0, os.path.join(dir_path, os.path.pardir))
    pw = importlib.import_module('%s.%s' % (
        os.path.basename(dir_path), 'write_param_card'))
    del sys.path[0]

    pw.ParamCardWriter('./param_card.dat', generic=True)
    print('write ./param_card.dat')



# type: ignore
################################################################################
#
# Copyright (c) 2012 The MadGraph5_aMC@NLO Development team and Contributors
#
# This file is a part of the MadGraph5_aMC@NLO project, an application which
# automatically generates Feynman diagrams and matrix elements for arbitrary
# high-energy processes in the Standard Model and beyond.
#
# It is subject to the MadGraph5_aMC@NLO license which should accompany this
# distribution.
#
# For more information, visit madgraph.phys.ucl.ac.be and amcatnlo.web.cern.ch
#
################################################################################
""" This part is not part of the UFO Model but only of MG5 suite. 
This files defines how the restrict card can be build automatically """

### Important Warning ###
# When you develop such file. Please cross check that they are NO
# unwanted simplification to your model. This can happen especially
# if a bunch of value are set to 1.0 by default. This YOUR responsability
# to check that you get the expected behavior
### Important Warning ###

from __future__ import absolute_import
import models.build_restriction_lib as build_restrict_lib
all_categories = []


first_category = build_restrict_lib.Category('sm customization')
all_categories.append(first_category)

first_category.add_options(name='diagonal ckm',  # name
                           default=True,          # default
                           rules=[('WOLFENSTEIN', [1], 0.0),
                                  ('WOLFENSTEIN', [2], 0.0),
                                  ('WOLFENSTEIN', [3], 0.0),
                                  ('WOLFENSTEIN', [4], 0.0)]
                           )

first_category.add_options(name='c mass = 0',  # name
                           default=True,        # default
                           rules=[('MASS', [4], 0.0),
                                  ('YUKAWA', [4], 0.0)]
                           )

first_category.add_options(name='b mass = 0',
                           default=False,
                           rules=[('MASS', [5], 0.0),
                                  ('YUKAWA', [5], 0.0)]
                           )

first_category.add_options(name='tau mass = 0',
                           default=False,
                           rules=[('MASS', [15], 0.0),
                                  ('YUKAWA', [15], 0.0)]
                           )

first_category.add_options(name='muon mass = 0',
                           default=True,
                           rules=[('MASS', [13], 0.0),
                                  ('YUKAWA', [13], 0.0)]
                           )

first_category.add_options(name='electron mass = 0',
                           default=True,
                           rules=[('MASS', [11], 0.0),
                                  ('YUKAWA', [11], 0.0)]
                           )



# type: ignore
# This file was automatically created by FeynRules 1.7.69
# Mathematica version: 8.0 for Mac OS X x86 (64-bit) (November 6, 2010)
# Date: Mon 1 Oct 2012 14:58:26


from __future__ import absolute_import

from .function_library import acsc, asec, complexconjugate, csc, im, re, sec
from .object_library import Lorentz, all_lorentz

try:
    import form_factors as ForFac
except ImportError:
    pass


UUS1 = Lorentz(name="UUS1", spins=[-1, -1, 1], structure="1")

UUV1 = Lorentz(name="UUV1", spins=[-1, -1, 3], structure="-P(3,2)")

SSS1 = Lorentz(name="SSS1", spins=[1, 1, 1], structure="1")

FFS1 = Lorentz(name="FFS1", spins=[2, 2, 1], structure="ProjM(1,2)")

FFS2 = Lorentz(name="FFS2", spins=[2, 2, 1], structure="ProjM(1,2) - ProjP(1,2)")

FFS3 = Lorentz(name="FFS3", spins=[2, 2, 1], structure="ProjP(1,2)")

FFS4 = Lorentz(name="FFS4", spins=[2, 2, 1], structure="ProjM(1,2) + ProjP(1,2)")

FFV1 = Lorentz(name="FFV1", spins=[2, 2, 3], structure="Gamma(3,1,2)")

FFV2 = Lorentz(name="FFV2", spins=[2, 2, 3], structure="Gamma(3,1,-1)*ProjM(-1,2)")

FFV3 = Lorentz(
    name="FFV3",
    spins=[2, 2, 3],
    structure="Gamma(3,1,-1)*ProjM(-1,2) - 2*Gamma(3,1,-1)*ProjP(-1,2)",
)

FFV4 = Lorentz(
    name="FFV4",
    spins=[2, 2, 3],
    structure="Gamma(3,1,-1)*ProjM(-1,2) + 2*Gamma(3,1,-1)*ProjP(-1,2)",
)

FFV5 = Lorentz(
    name="FFV5",
    spins=[2, 2, 3],
    structure="Gamma(3,1,-1)*ProjM(-1,2) + 4*Gamma(3,1,-1)*ProjP(-1,2)",
)

VSS1 = Lorentz(name="VSS1", spins=[3, 1, 1], structure="P(1,2) - P(1,3)")

VVS1 = Lorentz(name="VVS1", spins=[3, 3, 1], structure="Metric(1,2)")

VVV1 = Lorentz(
    name="VVV1",
    spins=[3, 3, 3],
    structure="P(3,1)*Metric(1,2) - P(3,2)*Metric(1,2) - P(2,1)*Metric(1,3) + P(2,3)*Metric(1,3) + P(1,2)*Metric(2,3) - P(1,3)*Metric(2,3)",
)

SSSS1 = Lorentz(name="SSSS1", spins=[1, 1, 1, 1], structure="1")

VVSS1 = Lorentz(name="VVSS1", spins=[3, 3, 1, 1], structure="Metric(1,2)")

VVVV1 = Lorentz(
    name="VVVV1",
    spins=[3, 3, 3, 3],
    structure="Metric(1,4)*Metric(2,3) - Metric(1,3)*Metric(2,4)",
)

VVVV2 = Lorentz(
    name="VVVV2",
    spins=[3, 3, 3, 3],
    structure="Metric(1,4)*Metric(2,3) + Metric(1,3)*Metric(2,4) - 2*Metric(1,2)*Metric(3,4)",
)

VVVV3 = Lorentz(
    name="VVVV3",
    spins=[3, 3, 3, 3],
    structure="Metric(1,4)*Metric(2,3) - Metric(1,2)*Metric(3,4)",
)

VVVV4 = Lorentz(
    name="VVVV4",
    spins=[3, 3, 3, 3],
    structure="Metric(1,3)*Metric(2,4) - Metric(1,2)*Metric(3,4)",
)

VVVV5 = Lorentz(
    name="VVVV5",
    spins=[3, 3, 3, 3],
    structure="Metric(1,4)*Metric(2,3) - (Metric(1,3)*Metric(2,4))/2. - (Metric(1,2)*Metric(3,4))/2.",
)



# type: ignore
# This file is part of the UFO.
#
# This file contains definitions for functions that
# are extensions of the cmath library, and correspond
# either to functions that are in cmath, but inconvenient
# to access from there (e.g. z.conjugate()),
# or functions that are simply not defined.
#
#
from __future__ import absolute_import
__date__ = "22 July 2010"
__author__ = "claude.duhr@durham.ac.uk"

from cmath import cos, sin, acos, asin
from .object_library import all_functions, Function

#
# shortcuts for functions from cmath
#

complexconjugate = Function(name='complexconjugate',
                            arguments=('z',),
                            expression='z.conjugate()')


re = Function(name='re',
              arguments=('z',),
              expression='z.real')

im = Function(name='im',
              arguments=('z',),
              expression='z.imag')


# New functions (trigonometric)

sec = Function(name='sec',
               arguments=('z',),
               expression='1./cos(z)')

asec = Function(name='asec',
                arguments=('z',),
                expression='acos(1./z)')

csc = Function(name='csc',
               arguments=('z',),
               expression='1./sin(z)')

acsc = Function(name='acsc',
                arguments=('z',),
                expression='asin(1./z)')



# type: ignore
# This file was automatically created by FeynRules 1.7.69
# Mathematica version: 8.0 for Mac OS X x86 (64-bit) (November 6, 2010)
# Date: Mon 1 Oct 2012 14:58:25


from __future__ import absolute_import
from .object_library import all_vertices, Vertex
from . import particles as P
from . import couplings as C
from . import lorentz as L
from . import parameters as Param
import itertools

for n_point_interaction in Param.N_POINT_INTERACTIONS:
    for i_inter, interaction in enumerate(sorted(list(set(tuple(sorted(inter)) for inter in itertools.product(*[list(range(Param.N_SCALARS)),]*n_point_interaction))))):
        # Remember that the constructor automatically adds the objects to all_vertices
        Vertex(name='V_%d_SCALAR_%s' % (n_point_interaction, ''.join('%d' % scalar_id for scalar_id in interaction)),
               particles=[P.scalar_particles[scalar_id]
                          for scalar_id in interaction],
               color=['1'],
               lorentz=[L.scalar_lorentz_structures[n_point_interaction]],
               couplings={(0, 0): C.SCALAR_COUPLING})



#
# Pretty horrenduous code ported from models/check_param_card.py in MadGraph5_aMC@NLO.
# It should be rewritten, but it will do for now
#

from __future__ import annotations
import math
import os
import re
from io import StringIO
from typing import Any
from ufo_model_loader.common import UFOModelLoaderError, logger, optionally_lower_external_parameter_name  # type: ignore
from ufo_model_loader.symbolica_processing import SYMBOLICA_AVAILABLE, expression_to_string  # type: ignore # nopep8
from pathlib import Path
from typing import TYPE_CHECKING, TextIO
from functools import cmp_to_key
if TYPE_CHECKING:
    from ufo_model_loader.model import Model, Particle  # type: ignore # nopep8
    from ufo_model_loader.model import Parameter as ModelParameter
pjoin = os.path.join

if SYMBOLICA_AVAILABLE and TYPE_CHECKING:
    from symbolica import Expression  # type: ignore


class InvalidParamCard(UFOModelLoaderError):
    """ An error class for invalid param_card """
    pass


class Parameter(object):
    """A class for a param_card parameter"""

    def __init__(self, block: str = '', lhacode: tuple[int, ...] | None = None, value: float = 0., comment: str = ''):
        """Init the parameter"""

        self.format: str = 'float'

        self.lhablock = block
        if lhacode:
            self.lhacode: tuple[int, ...] = lhacode
        else:
            self.lhacode: tuple[int, ...] = tuple([])  # type: ignore # nopep8
        self.value: float | str = value
        self.comment: str = comment

    def set_block(self, block: str):
        """ set the block name """

        self.lhablock = block

    def load_str(self, text: str):
        """ initialize the information from a str"""

        if '#' in text:
            data, self.comment = text.split('#', 1)
        else:
            data, self.comment = text, ""

        data = data.split()  # type: ignore # nopep8
        if any(d.startswith('scan') for d in data):
            position = [i for i, d in enumerate(
                data) if d.startswith('scan')][0]
            data = data[:position] + [' '.join(data[position:])]  # type: ignore # nopep8
        if not len(data):
            return
        try:
            self.lhacode = tuple([int(d) for d in data[:-1]])
        except Exception:
            self.lhacode = tuple([int(d) for d in data[:-1] if d.isdigit()])
            self.value = ' '.join(data[len(self.lhacode):])
        else:
            self.value = data[-1]

        # convert to number when possible
        try:
            self.value = float(self.value)
        except:
            self.format = 'str'
            pass
        else:
            if self.lhablock == 'modsel':
                self.format = 'int'
                self.value = int(self.value)

    def load_decay(self, text: str):
        """ initialize the decay information from a str"""

        if '#' in text:
            data, self.comment = text.split('#', 1)
        else:
            data, self.comment = text, ""

        if ']]>' in data:
            data = data.split(']]>', 1)[0]

        data = data.split()  # type: ignore # nopep8
        if not len(data):
            return
        lhacode = [int(d) for d in data[2:]]
        lhacode.sort()
        self.lhacode = tuple([len(lhacode)] + lhacode)

        self.value = float(data[0])
        self.format = 'decay_table'

    def __str__(self, precision: str | int = ''):
        """ return a SLAH string """

        format = self.format
        if self.format == 'float':
            try:
                _ = float(self.value)
            except:
                format = 'str'
        self.comment = self.comment.strip()
        if not precision:
            precision = 6

        self.comment = self.comment.strip()
        if format == 'float':
            if self.lhablock == 'decay' and not isinstance(self.value, str):
                return 'DECAY %s %.{0}e # %s'.format(precision) % (' '.join([str(d) for d in self.lhacode]), self.value, self.comment)
            elif self.lhablock == 'decay':
                return 'DECAY %s Auto # %s' % (' '.join([str(d) for d in self.lhacode]), self.comment)
            elif self.lhablock and self.lhablock.startswith('qnumbers'):
                return '      %s %i # %s' % (' '.join([str(d) for d in self.lhacode]), int(self.value), self.comment)
            else:
                return '      %s %.{0}e # %s'.format(precision) % (' '.join([str(d) for d in self.lhacode]), self.value, self.comment)
        elif format == 'int':
            return '      %s %i # %s' % (' '.join([str(d) for d in self.lhacode]), int(self.value), self.comment)
        elif format == 'str':
            if self.lhablock == 'decay':
                return 'DECAY %s %s # %s' % (' '.join([str(d) for d in self.lhacode]), self.value, self.comment)
            return '      %s %s # %s' % (' '.join([str(d) for d in self.lhacode]), self.value, self.comment)
        elif self.format == 'decay_table':
            return '      %e %s # %s' % (self.value, ' '.join([str(d) for d in self.lhacode]), self.comment)  # type: ignore # nopep8
        elif self.format == 'int':
            return '      %s %i # %s' % (' '.join([str(d) for d in self.lhacode]), int(self.value), self.comment)
        else:
            if self.lhablock == 'decay':
                return 'DECAY %s %d # %s' % (' '.join([str(d) for d in self.lhacode]), self.value, self.comment)  # type: ignore # nopep8
            else:
                return '      %s %d # %s' % (' '.join([str(d) for d in self.lhacode]), self.value, self.comment)  # type: ignore # nopep8


class Block(list[Parameter]):
    """ list of parameter """

    def __init__(self, name: str | None = None):
        if name:
            self.name: str = name.lower()
        else:
            self.name: str = ''  # type: ignore # nopep8
        self.scale = None
        self.comment = ''
        self.decay_table: dict[int, Block] = {}
        self.param_dict: dict[tuple[int, ...], Parameter] = {}
        list[Parameter].__init__(self)  # type: ignore # nopep8

    def get(self, lhacode_in: int | tuple[int, ...], default: Any = None) -> Parameter:
        """return the parameter associate to the lhacode"""
        if not self.param_dict:
            self.create_param_dict()

        if isinstance(lhacode_in, int):
            lhacode = (lhacode_in,)
        else:
            lhacode = lhacode_in  # type: ignore # nopep8

        try:
            return self.param_dict[tuple(lhacode)]
        except KeyError:
            if default is None:
                raise KeyError('id %s is not in %s' %
                               (tuple(lhacode), self.name))
            else:
                return Parameter(block=self.name, lhacode=tuple(lhacode), value=default,
                                 comment='not define')

    def remove(self, lhacode: tuple[int, ...]) -> Parameter:  # type: ignore
        """ remove a parameter """
        list[Parameter].remove(self, self.get(lhacode))  # type: ignore # nopep8
        # update the dictionary of key
        return self.param_dict.pop(tuple(lhacode))

    def __eq__(self, other: Block | str, prec: float = 1e-4):  # type: ignore
        """ """

        if isinstance(other, str) and ' ' not in other:
            return self.name.lower() == other.lower()

        if len(self) != len(other):
            return False

        return not any(abs(param.value-other.param_dict[key].value) > prec * abs(param.value)  # type: ignore
                       for key, param in self.param_dict.items() if isinstance(param.value, float))  # type: ignore

    def __ne__(self, other: Block, prec: float = 1e-4):  # type: ignore
        return not self.__eq__(other, prec)

    def append(self, obj: Parameter):

        if not hasattr(self, 'name'):  # can happen if loeaded from pickle
            self.__init__(obj.lhablock)  # type: ignore # nopep8
        assert not obj.lhablock or obj.lhablock == self.name

        # The following line seems/is stupid but allow to pickle/unpickle this object
        # this is important for madspin (in gridpack mode)
        if not hasattr(self, 'param_dict'):
            self.param_dict = {}

        if tuple(obj.lhacode) in self.param_dict:
            if self.param_dict[tuple(obj.lhacode)].value != obj.value:
                raise InvalidParamCard('%s %s is already define to %s impossible to assign %s' %
                                       (self.name, obj.lhacode, self.param_dict[tuple(obj.lhacode)].value, obj.value))
            return
        list[Parameter].append(self, obj)  # type: ignore # nopep8
        # update the dictionary of key
        self.param_dict[tuple(obj.lhacode)] = obj

    def create_param_dict(self):
        """create a link between the lhacode and the Parameter"""
        for param in self:
            self.param_dict[tuple(param.lhacode)] = param

        return self.param_dict

    def def_scale(self, scale: float):
        """ """
        self.scale = scale  # type: ignore # nopep8

    def load_str(self, text: str):
        "set inforamtion from the line"

        if '#' in text:
            data, self.comment = text.split('#', 1)
        else:
            data, self.comment = text, ""

        data = data.lower()
        data = data.split()  # type: ignore # nopep8
        self.name = data[1]  # the first part of data is model
        if len(data) == 3:
            if data[2].startswith('q='):
                # the last part should be of the form Q=
                self.scale = float(data[2][2:])  # type: ignore # nopep8
            elif self.name == 'qnumbers':
                self.name += ' %s' % data[2]
        elif len(data) == 4 and data[2] == 'q=':
            # the last part should be of the form Q=
            self.scale = float(data[3])  # type: ignore # nopep8

        return self

    def keys(self):
        """returns the list of id define in this blocks"""

        return [p.lhacode for p in self]

    def __str__(self, precision: str = ''):
        """ return a str in the SLAH format """

        text = """###################################""" + \
               """\n## INFORMATION FOR %s""" % self.name.upper() +\
               """\n###################################\n"""
        # special case for decay chain
        if self.name == 'decay':
            for param in self:
                pid = param.lhacode[0]
                param.set_block('decay')
                text += str(param) + '\n'
                if pid in self.decay_table:
                    text += str(self.decay_table[pid])+'\n'
            return text
        elif self.name.startswith('decay'):
            text = ''  # avoid block definition
        # general case
        elif not self.scale:
            text += 'BLOCK %s # %s\n' % (self.name.upper(), self.comment)
        else:
            text += 'BLOCK %s Q= %e # %s\n' % (self.name.upper(),
                                               self.scale, self.comment)

        text += '\n'.join([param.__str__(precision) for param in self])
        return text + '\n'


class ParamCard(dict[str, Block]):
    """ a param Card: list of Block """
    mp_prefix = 'MP__'

    header = \
        """######################################################################\n""" + \
        """## PARAM_CARD AUTOMATICALY GENERATED BY GAMMALOOP                 ####\n""" + \
        """######################################################################\n"""

    def __init__(self, input_path: str | ParamCard | None = None):
        dict[str, Block].__init__(self, {})   # type: ignore # nopep8
        self.order: list[Block] = []
        self.not_parsed_entry: list[str] = []

        if isinstance(input_path, ParamCard):
            i_p: ParamCard = input_path
            self.read(i_p.write())
            self.input_path = input_path.input_path   # type: ignore # nopep8
        else:
            self.input_path = input_path
            if input_path:
                self.read(input_path)

    def analyze_param_card(self, model: Model | None = None) -> tuple[dict[str, list[tuple[str, tuple[int, ...]]]], dict[tuple[str, tuple[int, ...]], str]]:
        """ Analyzes the comment of the parameter in the param_card and returns
        a dictionary with parameter names in values and the tuple (lhablock, id)
        in value as well as a dictionary for restricted values.
        WARNING: THIS FUNCTION RELIES ON THE FORMATTING OF THE COMMENT IN THE
        CARD TO FETCH THE PARAMETER NAME. This is mostly ok on the *_default.dat
        but typically dangerous on the user-defined card."""

        pname2block: dict[str, list[tuple[str, tuple[int, ...]]]] = {}
        restricted_value: dict[tuple[str, tuple[int, ...]], str] = {}

        for bname, block in self.items():
            for lha_id, param in block.param_dict.items():
                all_var = []
                comment = param.comment
                if model is not None:
                    model_parameter = model.get_parameter_from_lha_specification(
                        bname, lha_id)
                    if model_parameter is not None:
                        all_var = [optionally_lower_external_parameter_name(
                            model_parameter.name)]

                if len(all_var) == 0:
                    # treat merge parameter
                    if comment.strip().startswith('set of param :'):
                        all_var = list(re.findall(
                            r'''[^-]1\*(\w*)\b''', comment))
                    # just the variable name as comment
                    elif len(comment.split()) == 1:
                        all_var = [
                            optionally_lower_external_parameter_name(comment.strip())]
                    # either contraction or not formatted
                    else:
                        split = comment.split()
                        if len(split) > 2 and split[1] == ':':
                            # NO VAR associated
                            restricted_value[(bname, lha_id)
                                             ] = ' '.join(split[1:])
                        elif len(split) == 2:
                            if re.search(r'''\[[A-Z]\]eV\^''', split[1]):
                                all_var = [
                                    optionally_lower_external_parameter_name(comment.strip())]
                        elif len(split) >= 2 and split[1].startswith('('):
                            all_var = [optionally_lower_external_parameter_name(
                                split[0].strip())]
                        else:
                            if not bname.startswith('qnumbers') and comment != '':
                                logger.debug("information not recognized for %s %s : %s",
                                             bname, lha_id, comment)
                            continue

                for var in all_var:
                    var = optionally_lower_external_parameter_name(var)
                    if var in pname2block:
                        pname2block[var].append((bname, lha_id))
                    else:
                        pname2block[var] = [(bname, lha_id)]

        return (pname2block, restricted_value)

    def read(self, input_path: Any):
        """ read a card and full this object with the content of the card """

        if isinstance(input_path, str):
            if '\n' in input_path:
                input = StringIO(input_path)
            else:
                input = open(input_path)  # type: ignore # nopep8
        else:
            input = input_path  # Use for banner loading and test

        cur_block: Block | None | str = None
        for line in input:
            line = line.strip()
            if not line or line[0] == '#':
                continue
            line = line.lower()
            if line.startswith('block'):
                cur_block = Block()
                cur_block.load_str(line)
                self.append(cur_block)
                continue

            if line.startswith('decay'):
                if not self.has_block('decay'):
                    cur_block = Block('decay')
                    self.append(cur_block)
                else:
                    cur_block = self['decay']  # type: ignore
                param = Parameter()
                param.set_block(cur_block.name)  # type: ignore
                param.load_str(line[6:])
                cur_block.append(param)  # type: ignore
                continue

            if line.startswith('xsection') or cur_block == 'notparsed':
                cur_block = 'notparsed'
                self.not_parsed_entry.append(line)
                continue

            if cur_block is None:
                continue

            if cur_block.name == 'decay':  # type: ignore
                # This is a decay table
                decay_id = cur_block[-1].lhacode[0]  # type: ignore
                cur_block = Block('decay_table_%s' % decay_id)  # type: ignore
                self['decay'].decay_table[id] = cur_block  # type: ignore

            if cur_block.name.startswith('decay_table'):  # type: ignore
                param = Parameter()
                param.load_decay(line)
                try:
                    cur_block.append(param)  # type: ignore
                except InvalidParamCard:
                    pass
            else:
                param = Parameter()
                param.set_block(cur_block.name)  # type: ignore
                param.load_str(line)
                cur_block.append(param)  # type: ignore

        return self

    def __setitem__(self, name: str, value: Block) -> None:

        return dict[str, Block].__setitem__(self, name.lower(), value)  # type: ignore # nopep8

    def __getitem__(self, name: str) -> Block:
        return dict[str, Block].__getitem__(self, name.lower())  # type: ignore # nopep8

    def write(self, outpath: None | str | ParamCard = None, precision: str = '') -> str:
        """schedular for writing a card"""

        # order the block in a smart way
        blocks: list[Block] = self.order_block()
        text = self.header
        text += ''.join([block.__str__(precision) for block in blocks])
        text += '\n'
        text += '\n'.join(self.not_parsed_entry)
        if not outpath:
            return text
        elif isinstance(outpath, str):
            open(outpath, 'w').write(text)
            return ''
        else:
            return outpath.write(text)  # for test purpose

    def create_diff(self, new_card: ParamCard):
        """return a text file allowing to pass from this card to the new one
           via the set command"""

        diff = ''
        for blockname, block in self.items():
            for param in block:
                lhacode = param.lhacode
                value = param.value
                new_value = new_card[blockname].get(lhacode).value
                if isinstance(value, (float, int)) and isinstance(new_value, (float, int)):
                    same = (float(abs(value+new_value)) > 0. and not float(abs(value-new_value)/abs(
                        value+new_value)) < 1.0e-6) or (float(value) == 0. and float(new_value) == 0.)
                else:
                    same = value == new_value
                if same:
                    lhacode = ' '.join([str(i)
                                       for i in lhacode])  # type: ignore
                    diff += 'set param_card %s %s %s # orig: %s\n' % \
                        (blockname, lhacode, new_value, value)
        return diff

    def get_value(self, blockname: str, lhecode: int | tuple[int, ...], default: float | None = None) -> float | str:
        try:
            return self[blockname].get(lhecode).value
        except KeyError:
            if blockname == 'width':
                blockname = 'decay'
                return self.get_value(blockname, lhecode, default=default)
            elif default is not None:
                return default
            raise

    def convert_to_complex_mass_scheme(self):
        """ Convert this param_card to the convention used for the complex mass scheme:
        This includes, removing the Yukawa block if present and making sure the EW input
        scheme is (MZ, MW, aewm1). """

        # The yukawa block is irrelevant for the CMS models, we must remove them
        if self.has_block('yukawa'):
            # Notice that the last parameter removed will also remove the block.
            for lhacode in [param.lhacode for param in self['yukawa']]:
                self.remove_param('yukawa', lhacode)

        # Now fix the EW input scheme
        EW_input: dict[tuple[str, tuple[int, ...]], float | None] = {('sminputs', (1,)): None,  # type: ignore # nopep8
                                                                     ('sminputs', (2,)): None,
                                                                     ('mass', (23,)): None,
                                                                     ('mass', (24,)): None}
        for block, lhaid in EW_input.keys():
            try:
                EW_input[(block, lhaid)] = self[block].get(  # type: ignore
                    lhaid).value  # type: ignore
            except:
                pass

        # Now specify the missing values. We only support the following EW
        # input scheme:
        # (alpha, GF, MZ) input
        internal_param = [key for key,
                          value in EW_input.items() if value is None]
        if len(internal_param) == 0:
            # All parameters are already set, no need for modifications
            return

        if len(internal_param) != 1:
            raise InvalidParamCard(' The specified EW inputs has more than one' +
                                   ' unknown: [%s]' % (','.join([str(elem) for elem in internal_param])))

        if not internal_param[0] in [('mass', (24,)), ('sminputs', (2,)),
                                     ('sminputs', (1,))]:
            raise InvalidParamCard(' The only EW input scheme currently supported' +
                                   ' are those with either the W mass or GF left internal.')

        # Now if the Wmass is internal, then we must change the scheme
        if internal_param[0] == ('mass', (24,)):
            aewm1 = EW_input[('sminputs', (1,))]
            Gf = EW_input[('sminputs', (2,))]
            Mz = EW_input[('mass', (23,))]
            try:
                Mw = math.sqrt((Mz**2/2.0)+math.sqrt((Mz**4/4.0)-((  # type: ignore # nopep8
                              (1.0/aewm1)*math.pi*Mz**2)/(Gf*math.sqrt(2.0)))))  # type: ignore # nopep8
            except Exception as _exc:
                raise InvalidParamCard(
                    'The EW inputs 1/a_ew=%{aewm1:f}, Gf=%{Gf:f}, Mz=%{Mz:f} are inconsistent')
            self.remove_param('sminputs', (2,))
            self.add_param('mass', (24,), Mw, 'MW')

    def append(self, obj: Block):
        """add an object to this"""

        assert isinstance(obj, Block)
        self[obj.name] = obj
        if not obj.name.startswith('decay_table'):
            self.order.append(obj)

    def has_block(self, name: str) -> bool:
        return name in self

    def order_block(self) -> list[Block]:
        """ reorganize the block """
        return self.order

    def rename_blocks(self, name_dict: dict[str, str]):
        """ rename the blocks """

        for old_name, new_name in name_dict.items():
            self[new_name] = self.pop(old_name)
            self[new_name].name = new_name
            for param in self[new_name]:
                param.lhablock = new_name

    def remove_block(self, name: str):
        """ remove a blocks """
        assert len(self[name]) == 0
        [self.order.pop(i) for i, b in enumerate(self.order) if b.name == name]
        self.pop(name)

    def remove_param(self, block: str, lhacode: int | tuple[int, ...]) -> None:
        """ remove a parameter """
        if self.has_param(block, lhacode):  # type: ignore
            self[block].remove(lhacode)  # type: ignore
            if len(self[block]) == 0:
                self.remove_block(block)

    def has_param(self, block: str, lhacode: int):
        """check if param exists"""

        try:
            self[block].get(lhacode)
        except:
            return False
        else:
            return True

    def add_param(self, block: str, lha: int | tuple[int, ...], value: float, comment: str = ''):

        parameter = Parameter(block=block, lhacode=lha, value=value,  # type: ignore
                              comment=comment)
        try:
            new_block = self[block]
        except KeyError:
            # If the new block didn't exist yet
            new_block = Block(block)
            self.append(new_block)
        new_block.append(parameter)


class ParamCardWriter(object):

    header = \
        """######################################################################\n""" + \
        """## PARAM_CARD AUTOMATICALY GENERATED BY GAMMALOOP                   ##\n""" + \
        """######################################################################\n"""

    sm_pdg = [1, 2, 3, 4, 5, 6, 11, 12, 13, 13, 14, 15, 16, 21, 22, 23, 24, 25]
    qnumber_template = """Block QNUMBERS %(pdg)d  # %(name)s 
        1 %(charge)d  # 3 times electric charge
        2 %(spin)d  # number of spin states (2S+1)
        3 %(color)d  # colour rep (1: singlet, 3: triplet, 8: octet)
        4 %(antipart)d  # Particle/Antiparticle distinction (0=own anti)\n"""

    @staticmethod
    def write(output_path: Path, model: Model, generic: bool = True):
        """write a valid param_card.dat"""

        parameters = [p for p in model.get_external_parameters()
                      if p.lhablock is not None and p.lhacode is not None]

        dep_mass, dep_width = ParamCardWriter.define_not_dep_param(
            parameters, model.particles, generic)

        with open(output_path, 'w') as out:
            out.write(ParamCardWriter.header)
            ParamCardWriter.write_card(
                model.particles, parameters, dep_mass, dep_width, out, generic)

    @staticmethod
    def define_not_dep_param(parameters: list[ModelParameter], particles: list[Particle], generic: bool = True) -> tuple[list[tuple[Particle, ModelParameter]], list[tuple[Particle, ModelParameter]]]:
        """define self.dep_mass and self.dep_width in case that they are 
        requested in the param_card.dat"""

        dep_mass = []
        dep_width = []
        if generic:
            dep_mass = [(part, part.mass) for part in particles
                        if part.pdg_code > 0 and
                        part.mass not in parameters]
            dep_width = [(part, part.width) for part in particles
                         if part.pdg_code > 0 and
                         part.width not in parameters]

        return dep_mass, dep_width

    @staticmethod
    def order_param(obj1: ModelParameter, obj2: ModelParameter):
        """ order parameter of a given block """

        if obj1.lhacode is None or obj2.lhacode is None:
            return 0

        maxlen = min([len(obj1.lhacode), len(obj2.lhacode)])

        for i in range(maxlen):
            if obj1.lhacode[i] < obj2.lhacode[i]:
                return -1
            elif obj1.lhacode[i] == obj2.lhacode[i]:
                return 0
            else:
                return 1
        # identical up to the first finish
        if len(obj1.lhacode) > len(obj2.lhacode):
            return 1
        elif len(obj1.lhacode) == len(obj2.lhacode):
            return 0
        else:
            return -1

    @staticmethod
    def write_card(particles: list[Particle], parameters: list[ModelParameter], dep_mass: list[tuple[Particle, ModelParameter]], dep_width: list[tuple[Particle, ModelParameter]], out: TextIO, generic: bool = True):
        # list all lhablock
        all_lhablock = list(set(
            [param.lhablock for param in parameters if param.lhablock is not None]))

        # ordonate lhablock alphabeticaly
        all_lhablock = list(all_lhablock)  # type: ignore
        all_lhablock.sort(key=lambda x: x.lower())
        # put at the beginning SMINPUT + MASS + DECAY
        for name in ['DECAY', 'MASS', 'SMINPUTS']:
            if name in all_lhablock:
                all_lhablock.remove(name)
                all_lhablock.insert(0, name)

        for lhablock in all_lhablock:
            ParamCardWriter.write_block(lhablock, out)
            need_writing = [param for param in parameters if
                            param.lhablock == lhablock]
            need_writing.sort(key=cmp_to_key(ParamCardWriter.order_param))
            for param in need_writing:
                ParamCardWriter.write_param(param, lhablock, out)

            if generic:
                if lhablock in ['MASS', 'DECAY']:
                    ParamCardWriter.write_dep_param_block(
                        parameters, dep_mass, dep_width, lhablock, out)

        if generic:
            ParamCardWriter.write_qnumber(particles, out)

    @staticmethod
    def write_block(name: str, out: TextIO):
        """ write a comment for a block"""

        out.writelines(
            """\n###################################""" +
            """\n## INFORMATION FOR %s""" % name.upper() +
            """\n###################################\n"""
        )
        if name.upper() != 'DECAY':
            out.write("""Block %s \n""" % name)

    @staticmethod
    def write_param(param: ModelParameter, lhablock: str, out: TextIO):
        if param.lhablock is None or param.lhacode is None or param.value is None:
            return
        lhacode = ' '.join(['%3s' % key for key in param.lhacode])
        if lhablock != 'DECAY':
            text = """  %s %e # %s \n""" % (
                lhacode, complex(param.value).real, param.name)
        else:
            text = '''DECAY %s %e \n''' % (lhacode, complex(param.value).real)
        out.write(text)

    @staticmethod
    def write_dep_param_block(parameters: list[ModelParameter], dep_mass: list[tuple[Particle, ModelParameter]], dep_width: list[tuple[Particle, ModelParameter]], lhablock: str, out: TextIO):
        text = "##  Not dependent paramater.\n"
        text += "## Those values should be edited following analytical the \n"
        text += "## analytical expression. Some generator could simply ignore \n"
        text += "## those values and use the analytical expression\n"

        if lhablock == 'MASS':
            data = dep_mass
            prefix = " "
        else:
            data = dep_width
            prefix = "DECAY "
        for part, param in data:
            if param.value is None:
                continue
            text += """%s %s %f # %s : %s \n""" % (prefix, part.pdg_code,
                                                   param.value.real, part.name,
                                                   expression_to_string(param.expression))
        out.write(text)

    @staticmethod
    def write_qnumber(particles: list[Particle], out: TextIO):
        """ write qnumber """

        text = """#===========================================================\n"""
        text += """# QUANTUM NUMBERS OF NEW STATE(S) (NON SM PDG CODE)\n"""
        text += """#===========================================================\n\n"""

        for part in particles:
            if part.pdg_code in ParamCardWriter.sm_pdg or part.pdg_code < 0:
                continue
            text += ParamCardWriter.qnumber_template % {'pdg': part.pdg_code,
                                                        'name': part.name,
                                                        'charge': 3 * part.charge,
                                                        'spin': 2 * part.spin + 1,
                                                        'color': part.color,
                                                        'antipart': part.name != part.antiname and 1 or 0}

        out.write(text)



# type: ignore
# This file was automatically created by FeynRules 1.7.69
# Mathematica version: 8.0 for Mac OS X x86 (64-bit) (November 6, 2010)
# Date: Mon 1 Oct 2012 14:58:25


from __future__ import division
from __future__ import absolute_import
from .object_library import all_particles, Particle  # pylint: disable=unused-import
from . import parameters as Param

scalar_particles = []
for i_scalar in range(Param.N_SCALARS):
    scalar_particles.append(Particle(pdg_code=1000+i_scalar,
                                     name='scalar_%d' % i_scalar,
                                     antiname='scalar_%d' % i_scalar,
                                     spin=1,
                                     color=1,
                                     mass=Param.scalar_mass_parameters[i_scalar],
                                     width=Param.scalar_width_parameters[i_scalar],
                                     texname='\\phi_%d' % i_scalar,
                                     antitexname='\\phi_%d' % i_scalar,
                                     charge=0,
                                     GhostNumber=0,
                                     LeptonNumber=0,
                                     Y=0))



# This file was automatically created by FeynRules 1.7.69
# Mathematica version: 8.0 for Mac OS X x86 (64-bit) (November 6, 2010)
# Date: Mon 1 Oct 2012 14:58:25


from __future__ import division
from __future__ import absolute_import
from .object_library import all_particles, Particle
from . import parameters as Param

a = Particle(pdg_code=22,
             name='a',
             antiname='a',
             spin=3,
             color=1,
             mass=Param.ZERO,
             width=Param.ZERO,
             texname=r'{\gamma}',
             antitexname=r'{\gamma}',
             charge=0,
             GhostNumber=0,
             LeptonNumber=0,
             Y=0)

Z = Particle(pdg_code=23,
             name='Z',
             antiname='Z',
             spin=3,
             color=1,
             mass=Param.MZ,
             width=Param.WZ,
             texname=r'{Z}',
             antitexname=r'{Z}',
             charge=0,
             GhostNumber=0,
             LeptonNumber=0,
             Y=0)

W__plus__ = Particle(pdg_code=24,
                     name='W+',
                     antiname='W-',
                     spin=3,
                     color=1,
                     mass=Param.MW,
                     width=Param.WW,
                     texname=r'{W^+}',
                     antitexname=r'{W^-}',
                     charge=1,
                     GhostNumber=0,
                     LeptonNumber=0,
                     Y=0)

W__minus__ = W__plus__.anti()

g = Particle(pdg_code=21,
             name='g',
             antiname='g',
             spin=3,
             color=8,
             mass=Param.ZERO,
             width=Param.ZERO,
             texname=r'{g}',
             antitexname=r'{g}',
             charge=0,
             GhostNumber=0,
             LeptonNumber=0,
             Y=0)

ghA = Particle(pdg_code=9000001,
               name='ghA',
               antiname='ghA~',
               spin=-1,
               color=1,
               mass=Param.ZERO,
               width=Param.ZERO,
               texname=r'\tilde{\gamma}',
               antitexname=r'\overline{\tilde{\gamma}}',
               charge=0,
               GhostNumber=1,
               LeptonNumber=0,
               Y=0)

ghA__tilde__ = ghA.anti()

ghZ = Particle(pdg_code=9000002,
               name='ghZ',
               antiname='ghZ~',
               spin=-1,
               color=1,
               mass=Param.MZ,
               width=Param.WZ,
               texname=r'\tilde{Z}',
               antitexname=r'\bar{\tilde{Z}}',
               charge=0,
               GhostNumber=1,
               LeptonNumber=0,
               Y=0)

ghZ__tilde__ = ghZ.anti()

ghWp = Particle(pdg_code=9000003,
                name='ghWp',
                antiname='ghWp~',
                spin=-1,
                color=1,
                mass=Param.MW,
                width=Param.WW,
                texname=r'{\tilde{W}^+}',
                antitexname=r'{\bar{\tilde{W}}^+}',
                charge=1,
                GhostNumber=1,
                LeptonNumber=0,
                Y=0)

ghWp__tilde__ = ghWp.anti()

ghWm = Particle(pdg_code=9000004,
                name='ghWm',
                antiname='ghWm~',
                spin=-1,
                color=1,
                mass=Param.MW,
                width=Param.WW,
                texname=r'{\tilde{W}^-}',
                antitexname=r'{\bar{\tilde{W}}^-}',
                charge=-1,
                GhostNumber=1,
                LeptonNumber=0,
                Y=0)

ghWm__tilde__ = ghWm.anti()

ghG = Particle(pdg_code=9000005,
               name='ghG',
               antiname='ghG~',
               spin=-1,
               color=8,
               mass=Param.ZERO,
               width=Param.ZERO,
               texname=r'{\tilde{g}}',
               antitexname=r'{\bar{\tilde{g}}}',
               charge=0,
               GhostNumber=1,
               LeptonNumber=0,
               Y=0)

ghG__tilde__ = ghG.anti()

ve = Particle(pdg_code=12,
              name='ve',
              antiname='ve~',
              spin=2,
              color=1,
              mass=Param.ZERO,
              width=Param.ZERO,
              texname=r'{\nu_e}',
              antitexname=r'{\overline{\nu}_e}',
              charge=0,
              GhostNumber=0,
              LeptonNumber=1,
              Y=0)

ve__tilde__ = ve.anti()

vm = Particle(pdg_code=14,
              name='vm',
              antiname='vm~',
              spin=2,
              color=1,
              mass=Param.ZERO,
              width=Param.ZERO,
              texname=r'{\nu_\mu}',
              antitexname=r'{\overline{\nu}_\mu}',
              charge=0,
              GhostNumber=0,
              LeptonNumber=1,
              Y=0)

vm__tilde__ = vm.anti()

vt = Particle(pdg_code=16,
              name='vt',
              antiname='vt~',
              spin=2,
              color=1,
              mass=Param.ZERO,
              width=Param.ZERO,
              texname=r'{\nu_\tau}',
              antitexname=r'{\overline{\nu}_\tau}',
              charge=0,
              GhostNumber=0,
              LeptonNumber=1,
              Y=0)

vt__tilde__ = vt.anti()

u = Particle(pdg_code=2,
             name='u',
             antiname='u~',
             spin=2,
             color=3,
             mass=Param.ZERO,
             width=Param.ZERO,
             texname=r'{u}',
             antitexname=r'{\overline{u}}',
             charge=2/3,
             GhostNumber=0,
             LeptonNumber=0,
             Y=0)

u__tilde__ = u.anti()

c = Particle(pdg_code=4,
             name='c',
             antiname='c~',
             spin=2,
             color=3,
             mass=Param.MC,
             width=Param.ZERO,
             texname=r'{c}',
             antitexname=r'{\overline{c}}',
             charge=2/3,
             GhostNumber=0,
             LeptonNumber=0,
             Y=0)

c__tilde__ = c.anti()

t = Particle(pdg_code=6,
             name='t',
             antiname='t~',
             spin=2,
             color=3,
             mass=Param.MT,
             width=Param.WT,
             texname=r'{t}',
             antitexname=r'{\overline{t}}',
             charge=2/3,
             GhostNumber=0,
             LeptonNumber=0,
             Y=0)

t__tilde__ = t.anti()

d = Particle(pdg_code=1,
             name='d',
             antiname='d~',
             spin=2,
             color=3,
             mass=Param.ZERO,
             width=Param.ZERO,
             texname=r'{d}',
             antitexname=r'{\overline{d}}',
             charge=-1/3,
             GhostNumber=0,
             LeptonNumber=0,
             Y=0)

d__tilde__ = d.anti()

s = Particle(pdg_code=3,
             name='s',
             antiname='s~',
             spin=2,
             color=3,
             mass=Param.ZERO,
             width=Param.ZERO,
             texname=r'{s}',
             antitexname=r'{\overline{s}}',
             charge=-1/3,
             GhostNumber=0,
             LeptonNumber=0,
             Y=0)

s__tilde__ = s.anti()

b = Particle(pdg_code=5,
             name='b',
             antiname='b~',
             spin=2,
             color=3,
             mass=Param.MB,
             width=Param.ZERO,
             texname=r'{b}',
             antitexname=r'{\overline{b}}',
             charge=-1/3,
             GhostNumber=0,
             LeptonNumber=0,
             Y=0)

b__tilde__ = b.anti()

H = Particle(pdg_code=25,
             name='H',
             antiname='H',
             spin=1,
             color=1,
             mass=Param.MH,
             width=Param.WH,
             texname=r'{H}',
             antitexname=r'{H}',
             charge=0,
             GhostNumber=0,
             LeptonNumber=0,
             Y=0)

G0 = Particle(pdg_code=250,
              name='G0',
              antiname='G0',
              spin=1,
              color=1,
              mass=Param.MZ,
              width=Param.WZ,
              texname=r'{G_0}',
              antitexname=r'{G_0}',
              GoldstoneBoson=True,
              charge=0,
              GhostNumber=0,
              LeptonNumber=0,
              Y=0)

G__plus__ = Particle(pdg_code=251,
                     name='G+',
                     antiname='G-',
                     spin=1,
                     color=1,
                     mass=Param.MW,
                     width=Param.WW,
                     texname=r'{G^+}',
                     antitexname=r'{G^-}',
                     GoldstoneBoson=True,
                     charge=1,
                     GhostNumber=0,
                     LeptonNumber=0,
                     Y=0)

G__minus__ = G__plus__.anti()

e__minus__ = Particle(pdg_code=11,
                      name='e-',
                      antiname='e+',
                      spin=2,
                      color=1,
                      mass=Param.Me,
                      width=Param.ZERO,
                      texname=r'{e^-}',
                      antitexname=r'{e^+}',
                      charge=-1,
                      GhostNumber=0,
                      LeptonNumber=1,
                      Y=0)

e__plus__ = e__minus__.anti()

mu__minus__ = Particle(pdg_code=13,
                       name='mu-',
                       antiname='mu+',
                       spin=2,
                       color=1,
                       mass=Param.MM,
                       width=Param.ZERO,
                       texname=r'{\mu^-}',
                       antitexname=r'{\mu^+}',
                       charge=-1,
                       GhostNumber=0,
                       LeptonNumber=1,
                       Y=0)

mu__plus__ = mu__minus__.anti()

ta__minus__ = Particle(pdg_code=15,
                       name='ta-',
                       antiname='ta+',
                       spin=2,
                       color=1,
                       mass=Param.MTA,
                       width=Param.WTau,
                       texname=r'{\tau^-}',
                       antitexname=r'{\tau^+}',
                       charge=-1,
                       GhostNumber=0,
                       LeptonNumber=1,
                       Y=0)

ta__plus__ = ta__minus__.anti()



# This file was automatically created by FeynRules 1.7.69
# Mathematica version: 8.0 for Mac OS X x86 (64-bit) (November 6, 2010)
# Date: Mon 1 Oct 2012 14:58:26


from __future__ import absolute_import
from .object_library import all_couplings, Coupling

from .function_library import complexconjugate, re, im, csc, sec, acsc, asec



GC_1 = Coupling(name = 'GC_1',
                value = '-(ee*complex(0,1))/3.',
                order = {'QED':1})

GC_2 = Coupling(name = 'GC_2',
                value = '(2*ee*complex(0,1))/3.',
                order = {'QED':1})

GC_3 = Coupling(name = 'GC_3',
                value = '-(ee*complex(0,1))',
                order = {'QED':1})

GC_4 = Coupling(name = 'GC_4',
                value = 'ee*complex(0,1)',
                order = {'QED':1})

GC_5 = Coupling(name = 'GC_5',
                value = 'ee**2*complex(0,1)',
                order = {'QED':2})

GC_6 = Coupling(name = 'GC_6',
                value = '2*ee**2*complex(0,1)',
                order = {'QED':2})

GC_7 = Coupling(name = 'GC_7',
                value = '-ee**2/(2.*cw)',
                order = {'QED':2})

GC_8 = Coupling(name = 'GC_8',
                value = '(ee**2*complex(0,1))/(2.*cw)',
                order = {'QED':2})

GC_9 = Coupling(name = 'GC_9',
                value = 'ee**2/(2.*cw)',
                order = {'QED':2})

GC_10 = Coupling(name = 'GC_10',
                 value = '-G',
                 order = {'QCD':1})

GC_11 = Coupling(name = 'GC_11',
                 value = 'complex(0,1)*G',
                 order = {'QCD':1})

GC_12 = Coupling(name = 'GC_12',
                 value = 'complex(0,1)*G**2',
                 order = {'QCD':2})

GC_13 = Coupling(name = 'GC_13',
                 value = 'I1x31',
                 order = {'QED':1})

GC_14 = Coupling(name = 'GC_14',
                 value = 'I1x32',
                 order = {'QED':1})

GC_15 = Coupling(name = 'GC_15',
                 value = 'I1x33',
                 order = {'QED':1})

GC_16 = Coupling(name = 'GC_16',
                 value = '-I2x12',
                 order = {'QED':1})

GC_17 = Coupling(name = 'GC_17',
                 value = '-I2x13',
                 order = {'QED':1})

GC_18 = Coupling(name = 'GC_18',
                 value = '-I2x22',
                 order = {'QED':1})

GC_19 = Coupling(name = 'GC_19',
                 value = '-I2x23',
                 order = {'QED':1})

GC_20 = Coupling(name = 'GC_20',
                 value = '-I2x32',
                 order = {'QED':1})

GC_21 = Coupling(name = 'GC_21',
                 value = '-I2x33',
                 order = {'QED':1})

GC_22 = Coupling(name = 'GC_22',
                 value = 'I3x21',
                 order = {'QED':1})

GC_23 = Coupling(name = 'GC_23',
                 value = 'I3x22',
                 order = {'QED':1})

GC_24 = Coupling(name = 'GC_24',
                 value = 'I3x23',
                 order = {'QED':1})

GC_25 = Coupling(name = 'GC_25',
                 value = 'I3x31',
                 order = {'QED':1})

GC_26 = Coupling(name = 'GC_26',
                 value = 'I3x32',
                 order = {'QED':1})

GC_27 = Coupling(name = 'GC_27',
                 value = 'I3x33',
                 order = {'QED':1})

GC_28 = Coupling(name = 'GC_28',
                 value = '-I4x13',
                 order = {'QED':1})

GC_29 = Coupling(name = 'GC_29',
                 value = '-I4x23',
                 order = {'QED':1})

GC_30 = Coupling(name = 'GC_30',
                 value = '-I4x33',
                 order = {'QED':1})

GC_31 = Coupling(name = 'GC_31',
                 value = '-2*complex(0,1)*lam',
                 order = {'QED':2})

GC_32 = Coupling(name = 'GC_32',
                 value = '-4*complex(0,1)*lam',
                 order = {'QED':2})

GC_33 = Coupling(name = 'GC_33',
                 value = '-6*complex(0,1)*lam',
                 order = {'QED':2})

GC_34 = Coupling(name = 'GC_34',
                 value = '(ee**2*complex(0,1))/(2.*sw**2)',
                 order = {'QED':2})

GC_35 = Coupling(name = 'GC_35',
                 value = '-((ee**2*complex(0,1))/sw**2)',
                 order = {'QED':2})

GC_36 = Coupling(name = 'GC_36',
                 value = '(cw**2*ee**2*complex(0,1))/sw**2',
                 order = {'QED':2})

GC_37 = Coupling(name = 'GC_37',
                 value = '-ee/(2.*sw)',
                 order = {'QED':1})

GC_38 = Coupling(name = 'GC_38',
                 value = '-(ee*complex(0,1))/(2.*sw)',
                 order = {'QED':1})

GC_39 = Coupling(name = 'GC_39',
                 value = '(ee*complex(0,1))/(2.*sw)',
                 order = {'QED':1})

GC_40 = Coupling(name = 'GC_40',
                 value = '(ee*complex(0,1))/(sw*cmath.sqrt(2))',
                 order = {'QED':1})

GC_41 = Coupling(name = 'GC_41',
                 value = '(CKM1x1*ee*complex(0,1))/(sw*cmath.sqrt(2))',
                 order = {'QED':1})

GC_42 = Coupling(name = 'GC_42',
                 value = '(CKM1x2*ee*complex(0,1))/(sw*cmath.sqrt(2))',
                 order = {'QED':1})

GC_43 = Coupling(name = 'GC_43',
                 value = '(CKM1x3*ee*complex(0,1))/(sw*cmath.sqrt(2))',
                 order = {'QED':1})

GC_44 = Coupling(name = 'GC_44',
                 value = '(CKM2x1*ee*complex(0,1))/(sw*cmath.sqrt(2))',
                 order = {'QED':1})

GC_45 = Coupling(name = 'GC_45',
                 value = '(CKM2x2*ee*complex(0,1))/(sw*cmath.sqrt(2))',
                 order = {'QED':1})

GC_46 = Coupling(name = 'GC_46',
                 value = '(CKM2x3*ee*complex(0,1))/(sw*cmath.sqrt(2))',
                 order = {'QED':1})

GC_47 = Coupling(name = 'GC_47',
                 value = '(CKM3x1*ee*complex(0,1))/(sw*cmath.sqrt(2))',
                 order = {'QED':1})

GC_48 = Coupling(name = 'GC_48',
                 value = '(CKM3x2*ee*complex(0,1))/(sw*cmath.sqrt(2))',
                 order = {'QED':1})

GC_49 = Coupling(name = 'GC_49',
                 value = '(CKM3x3*ee*complex(0,1))/(sw*cmath.sqrt(2))',
                 order = {'QED':1})

GC_50 = Coupling(name = 'GC_50',
                 value = '-(cw*ee*complex(0,1))/(2.*sw)',
                 order = {'QED':1})

GC_51 = Coupling(name = 'GC_51',
                 value = '(cw*ee*complex(0,1))/(2.*sw)',
                 order = {'QED':1})

GC_52 = Coupling(name = 'GC_52',
                 value = '-((cw*ee*complex(0,1))/sw)',
                 order = {'QED':1})

GC_53 = Coupling(name = 'GC_53',
                 value = '(cw*ee*complex(0,1))/sw',
                 order = {'QED':1})

GC_54 = Coupling(name = 'GC_54',
                 value = '-ee**2/(2.*sw)',
                 order = {'QED':2})

GC_55 = Coupling(name = 'GC_55',
                 value = '-(ee**2*complex(0,1))/(2.*sw)',
                 order = {'QED':2})

GC_56 = Coupling(name = 'GC_56',
                 value = 'ee**2/(2.*sw)',
                 order = {'QED':2})

GC_57 = Coupling(name = 'GC_57',
                 value = '(-2*cw*ee**2*complex(0,1))/sw',
                 order = {'QED':2})

GC_58 = Coupling(name = 'GC_58',
                 value = '-(ee*complex(0,1)*sw)/(6.*cw)',
                 order = {'QED':1})

GC_59 = Coupling(name = 'GC_59',
                 value = '(ee*complex(0,1)*sw)/(2.*cw)',
                 order = {'QED':1})

GC_60 = Coupling(name = 'GC_60',
                 value = '-(cw*ee)/(2.*sw) - (ee*sw)/(2.*cw)',
                 order = {'QED':1})

GC_61 = Coupling(name = 'GC_61',
                 value = '-(cw*ee*complex(0,1))/(2.*sw) + (ee*complex(0,1)*sw)/(2.*cw)',
                 order = {'QED':1})

GC_62 = Coupling(name = 'GC_62',
                 value = '(cw*ee*complex(0,1))/(2.*sw) + (ee*complex(0,1)*sw)/(2.*cw)',
                 order = {'QED':1})

GC_63 = Coupling(name = 'GC_63',
                 value = '(cw*ee**2*complex(0,1))/sw - (ee**2*complex(0,1)*sw)/cw',
                 order = {'QED':2})

GC_64 = Coupling(name = 'GC_64',
                 value = '-(ee**2*complex(0,1)) + (cw**2*ee**2*complex(0,1))/(2.*sw**2) + (ee**2*complex(0,1)*sw**2)/(2.*cw**2)',
                 order = {'QED':2})

GC_65 = Coupling(name = 'GC_65',
                 value = 'ee**2*complex(0,1) + (cw**2*ee**2*complex(0,1))/(2.*sw**2) + (ee**2*complex(0,1)*sw**2)/(2.*cw**2)',
                 order = {'QED':2})

GC_66 = Coupling(name = 'GC_66',
                 value = '-(ee**2*vev)/(2.*cw)',
                 order = {'QED':1})

GC_67 = Coupling(name = 'GC_67',
                 value = '(ee**2*vev)/(2.*cw)',
                 order = {'QED':1})

GC_68 = Coupling(name = 'GC_68',
                 value = '-2*complex(0,1)*lam*vev',
                 order = {'QED':1})

GC_69 = Coupling(name = 'GC_69',
                 value = '-6*complex(0,1)*lam*vev',
                 order = {'QED':1})

GC_70 = Coupling(name = 'GC_70',
                 value = '-(ee**2*vev)/(4.*sw**2)',
                 order = {'QED':1})

GC_71 = Coupling(name = 'GC_71',
                 value = '-(ee**2*complex(0,1)*vev)/(4.*sw**2)',
                 order = {'QED':1})

GC_72 = Coupling(name = 'GC_72',
                 value = '(ee**2*complex(0,1)*vev)/(2.*sw**2)',
                 order = {'QED':1})

GC_73 = Coupling(name = 'GC_73',
                 value = '(ee**2*vev)/(4.*sw**2)',
                 order = {'QED':1})

GC_74 = Coupling(name = 'GC_74',
                 value = '-(ee**2*vev)/(2.*sw)',
                 order = {'QED':1})

GC_75 = Coupling(name = 'GC_75',
                 value = '(ee**2*vev)/(2.*sw)',
                 order = {'QED':1})

GC_76 = Coupling(name = 'GC_76',
                 value = '-(ee**2*vev)/(4.*cw) - (cw*ee**2*vev)/(4.*sw**2)',
                 order = {'QED':1})

GC_77 = Coupling(name = 'GC_77',
                 value = '(ee**2*vev)/(4.*cw) - (cw*ee**2*vev)/(4.*sw**2)',
                 order = {'QED':1})

GC_78 = Coupling(name = 'GC_78',
                 value = '-(ee**2*vev)/(4.*cw) + (cw*ee**2*vev)/(4.*sw**2)',
                 order = {'QED':1})

GC_79 = Coupling(name = 'GC_79',
                 value = '(ee**2*vev)/(4.*cw) + (cw*ee**2*vev)/(4.*sw**2)',
                 order = {'QED':1})

GC_80 = Coupling(name = 'GC_80',
                 value = '-(ee**2*complex(0,1)*vev)/2. - (cw**2*ee**2*complex(0,1)*vev)/(4.*sw**2) - (ee**2*complex(0,1)*sw**2*vev)/(4.*cw**2)',
                 order = {'QED':1})

GC_81 = Coupling(name = 'GC_81',
                 value = 'ee**2*complex(0,1)*vev + (cw**2*ee**2*complex(0,1)*vev)/(2.*sw**2) + (ee**2*complex(0,1)*sw**2*vev)/(2.*cw**2)',
                 order = {'QED':1})

GC_82 = Coupling(name = 'GC_82',
                 value = '-(yb/cmath.sqrt(2))',
                 order = {'QED':1})

GC_83 = Coupling(name = 'GC_83',
                 value = '-((complex(0,1)*yb)/cmath.sqrt(2))',
                 order = {'QED':1})

GC_84 = Coupling(name = 'GC_84',
                 value = '-((complex(0,1)*yc)/cmath.sqrt(2))',
                 order = {'QED':1})

GC_85 = Coupling(name = 'GC_85',
                 value = 'yc/cmath.sqrt(2)',
                 order = {'QED':1})

GC_86 = Coupling(name = 'GC_86',
                 value = '-ye',
                 order = {'QED':1})

GC_87 = Coupling(name = 'GC_87',
                 value = 'ye',
                 order = {'QED':1})

GC_88 = Coupling(name = 'GC_88',
                 value = '-(ye/cmath.sqrt(2))',
                 order = {'QED':1})

GC_89 = Coupling(name = 'GC_89',
                 value = '-((complex(0,1)*ye)/cmath.sqrt(2))',
                 order = {'QED':1})

GC_90 = Coupling(name = 'GC_90',
                 value = '-ym',
                 order = {'QED':1})

GC_91 = Coupling(name = 'GC_91',
                 value = 'ym',
                 order = {'QED':1})

GC_92 = Coupling(name = 'GC_92',
                 value = '-(ym/cmath.sqrt(2))',
                 order = {'QED':1})

GC_93 = Coupling(name = 'GC_93',
                 value = '-((complex(0,1)*ym)/cmath.sqrt(2))',
                 order = {'QED':1})

GC_94 = Coupling(name = 'GC_94',
                 value = '-((complex(0,1)*yt)/cmath.sqrt(2))',
                 order = {'QED':1})

GC_95 = Coupling(name = 'GC_95',
                 value = 'yt/cmath.sqrt(2)',
                 order = {'QED':1})

GC_96 = Coupling(name = 'GC_96',
                 value = '-ytau',
                 order = {'QED':1})

GC_97 = Coupling(name = 'GC_97',
                 value = 'ytau',
                 order = {'QED':1})

GC_98 = Coupling(name = 'GC_98',
                 value = '-(ytau/cmath.sqrt(2))',
                 order = {'QED':1})

GC_99 = Coupling(name = 'GC_99',
                 value = '-((complex(0,1)*ytau)/cmath.sqrt(2))',
                 order = {'QED':1})

GC_100 = Coupling(name = 'GC_100',
                  value = '(ee*complex(0,1)*complexconjugate(CKM1x1))/(sw*cmath.sqrt(2))',
                  order = {'QED':1})

GC_101 = Coupling(name = 'GC_101',
                  value = '(ee*complex(0,1)*complexconjugate(CKM1x2))/(sw*cmath.sqrt(2))',
                  order = {'QED':1})

GC_102 = Coupling(name = 'GC_102',
                  value = '(ee*complex(0,1)*complexconjugate(CKM1x3))/(sw*cmath.sqrt(2))',
                  order = {'QED':1})

GC_103 = Coupling(name = 'GC_103',
                  value = '(ee*complex(0,1)*complexconjugate(CKM2x1))/(sw*cmath.sqrt(2))',
                  order = {'QED':1})

GC_104 = Coupling(name = 'GC_104',
                  value = '(ee*complex(0,1)*complexconjugate(CKM2x2))/(sw*cmath.sqrt(2))',
                  order = {'QED':1})

GC_105 = Coupling(name = 'GC_105',
                  value = '(ee*complex(0,1)*complexconjugate(CKM2x3))/(sw*cmath.sqrt(2))',
                  order = {'QED':1})

GC_106 = Coupling(name = 'GC_106',
                  value = '(ee*complex(0,1)*complexconjugate(CKM3x1))/(sw*cmath.sqrt(2))',
                  order = {'QED':1})

GC_107 = Coupling(name = 'GC_107',
                  value = '(ee*complex(0,1)*complexconjugate(CKM3x2))/(sw*cmath.sqrt(2))',
                  order = {'QED':1})

GC_108 = Coupling(name = 'GC_108',
                  value = '(ee*complex(0,1)*complexconjugate(CKM3x3))/(sw*cmath.sqrt(2))',
                  order = {'QED':1})




# This file was automatically created by FeynRules 1.7.69
# Mathematica version: 8.0 for Mac OS X x86 (64-bit) (November 6, 2010)
# Date: Mon 1 Oct 2012 14:58:25


from __future__ import absolute_import
from .object_library import all_parameters, Parameter

from .function_library import complexconjugate, re, im, csc, sec, acsc, asec

# This is a default parameter object representing 0.
ZERO = Parameter(name='ZERO',
                 nature='internal',
                 type='real',
                 value=0.0,
                 texname='0')

# User-defined parameters.
aEWM1 = Parameter(name='aEWM1',
                  nature='external',
                  type='real',
                  value=132.50698,
                  texname='\\text{aEWM1}',
                  lhablock='SMINPUTS',
                  lhacode=[1])

Gf = Parameter(name='Gf',
               nature='external',
               type='real',
               value=0.0000116639,
               texname='G_f',
               lhablock='SMINPUTS',
               lhacode=[2])

aS = Parameter(name='aS',
               nature='external',
               type='real',
               value=0.118,
               texname='\\alpha _s',
               lhablock='SMINPUTS',
               lhacode=[3])

lamWS = Parameter(name='lamWS',
                  nature='external',
                  type='real',
                  value=0.2253,
                  texname='\\text{lamWS}',
                  lhablock='Wolfenstein',
                  lhacode=[1])

AWS = Parameter(name='AWS',
                nature='external',
                type='real',
                value=0.808,
                texname='\\text{AWS}',
                lhablock='Wolfenstein',
                lhacode=[2])

rhoWS = Parameter(name='rhoWS',
                  nature='external',
                  type='real',
                  value=0.132,
                  texname='\\text{rhoWS}',
                  lhablock='Wolfenstein',
                  lhacode=[3])

etaWS = Parameter(name='etaWS',
                  nature='external',
                  type='real',
                  value=0.341,
                  texname='\\text{etaWS}',
                  lhablock='Wolfenstein',
                  lhacode=[4])

ymc = Parameter(name='ymc',
                nature='external',
                type='real',
                value=1.27,
                texname='\\text{ymc}',
                lhablock='YUKAWA',
                lhacode=[4])

ymb = Parameter(name='ymb',
                nature='external',
                type='real',
                value=4.2,
                texname='\\text{ymb}',
                lhablock='YUKAWA',
                lhacode=[5])

ymt = Parameter(name='ymt',
                nature='external',
                type='real',
                value=164.5,
                texname='\\text{ymt}',
                lhablock='YUKAWA',
                lhacode=[6])

yme = Parameter(name='yme',
                nature='external',
                type='real',
                value=0.000511,
                texname='\\text{yme}',
                lhablock='YUKAWA',
                lhacode=[11])

ymm = Parameter(name='ymm',
                nature='external',
                type='real',
                value=0.10566,
                texname='\\text{ymm}',
                lhablock='YUKAWA',
                lhacode=[13])

ymtau = Parameter(name='ymtau',
                  nature='external',
                  type='real',
                  value=1.777,
                  texname='\\text{ymtau}',
                  lhablock='YUKAWA',
                  lhacode=[15])

MZ = Parameter(name='MZ',
               nature='external',
               type='real',
               value=91.188,
               texname='\\text{MZ}',
               lhablock='MASS',
               lhacode=[23])

MC = Parameter(name='MC',
               nature='external',
               type='real',
               value=1.27,
               texname='\\text{MC}',
               lhablock='MASS',
               lhacode=[4])

MT = Parameter(name='MT',
               nature='external',
               type='real',
               value=173.,
               texname='\\text{MT}',
               lhablock='MASS',
               lhacode=[6])

MB = Parameter(name='MB',
               nature='external',
               type='real',
               value=4.7,
               texname='\\text{MB}',
               lhablock='MASS',
               lhacode=[5])

MH = Parameter(name='MH',
               nature='external',
               type='real',
               value=125.,
               texname='\\text{MH}',
               lhablock='MASS',
               lhacode=[25])

Me = Parameter(name='Me',
               nature='external',
               type='real',
               value=0.000511,
               texname='\\text{Me}',
               lhablock='MASS',
               lhacode=[11])

MM = Parameter(name='MM',
               nature='external',
               type='real',
               value=0.10566,
               texname='\\text{MM}',
               lhablock='MASS',
               lhacode=[13])

MTA = Parameter(name='MTA',
                nature='external',
                type='real',
                value=1.777,
                texname='\\text{MTA}',
                lhablock='MASS',
                lhacode=[15])

WZ = Parameter(name='WZ',
               nature='external',
               type='real',
               value=2.44140351,
               texname='\\text{WZ}',
               lhablock='DECAY',
               lhacode=[23])

WW = Parameter(name='WW',
               nature='external',
               type='real',
               value=2.04759951,
               texname='\\text{WW}',
               lhablock='DECAY',
               lhacode=[24])

WT = Parameter(name='WT',
               nature='external',
               type='real',
               value=1.50833649,
               texname='\\text{WT}',
               lhablock='DECAY',
               lhacode=[6])

WH = Parameter(name='WH',
               nature='external',
               type='real',
               value=6.38233934e-03,
               texname='\\text{WH}',
               lhablock='DECAY',
               lhacode=[25])

WTau = Parameter(name='WTau',
                 nature='external',
                 type='real',
                 value=2.27e-12,
                 texname='\\text{WTau}',
                 lhablock='DECAY',
                 lhacode=[15])

CKM1x1 = Parameter(name='CKM1x1',
                   nature='internal',
                   type='complex',
                   value='1 - lamWS**2/2.',
                   texname='\\text{CKM1x1}')

CKM1x2 = Parameter(name='CKM1x2',
                   nature='internal',
                   type='complex',
                   value='lamWS',
                   texname='\\text{CKM1x2}')

CKM1x3 = Parameter(name='CKM1x3',
                   nature='internal',
                   type='complex',
                   value='AWS*lamWS**3*(-(etaWS*complex(0,1)) + rhoWS)',
                   texname='\\text{CKM1x3}')

CKM2x1 = Parameter(name='CKM2x1',
                   nature='internal',
                   type='complex',
                   value='-lamWS',
                   texname='\\text{CKM2x1}')

CKM2x2 = Parameter(name='CKM2x2',
                   nature='internal',
                   type='complex',
                   value='1 - lamWS**2/2.',
                   texname='\\text{CKM2x2}')

CKM2x3 = Parameter(name='CKM2x3',
                   nature='internal',
                   type='complex',
                   value='AWS*lamWS**2',
                   texname='\\text{CKM2x3}')

CKM3x1 = Parameter(name='CKM3x1',
                   nature='internal',
                   type='complex',
                   value='AWS*lamWS**3*(1 - etaWS*complex(0,1) - rhoWS)',
                   texname='\\text{CKM3x1}')

CKM3x2 = Parameter(name='CKM3x2',
                   nature='internal',
                   type='complex',
                   value='-(AWS*lamWS**2)',
                   texname='\\text{CKM3x2}')

CKM3x3 = Parameter(name='CKM3x3',
                   nature='internal',
                   type='complex',
                   value='1',
                   texname='\\text{CKM3x3}')

aEW = Parameter(name='aEW',
                nature='internal',
                type='real',
                value='1/aEWM1',
                texname='\\alpha _{\\text{EW}}')

G = Parameter(name='G',
              nature='internal',
              type='real',
              value='2*cmath.sqrt(aS)*cmath.sqrt(cmath.pi)',
              texname='G')

MW = Parameter(name='MW',
               nature='internal',
               type='real',
               value='cmath.sqrt(MZ**2/2. + cmath.sqrt(MZ**4/4. - (aEW*cmath.pi*MZ**2)/(Gf*cmath.sqrt(2))))',
               texname='M_W')

ee = Parameter(name='ee',
               nature='internal',
               type='real',
               value='2*cmath.sqrt(aEW)*cmath.sqrt(cmath.pi)',
               texname='e')

sw2 = Parameter(name='sw2',
                nature='internal',
                type='real',
                value='1 - MW**2/MZ**2',
                texname='\\text{sw2}')

cw = Parameter(name='cw',
               nature='internal',
               type='real',
               value='cmath.sqrt(1 - sw2)',
               texname='c_w')

sw = Parameter(name='sw',
               nature='internal',
               type='real',
               value='cmath.sqrt(sw2)',
               texname='s_w')

g1 = Parameter(name='g1',
               nature='internal',
               type='real',
               value='ee/cw',
               texname='g_1')

gw = Parameter(name='gw',
               nature='internal',
               type='real',
               value='ee/sw',
               texname='g_w')

vev = Parameter(name='vev',
                nature='internal',
                type='real',
                value='(2*MW*sw)/ee',
                texname='\\text{vev}')

lam = Parameter(name='lam',
                nature='internal',
                type='real',
                value='MH**2/(2.*vev**2)',
                texname='\\text{lam}')

yb = Parameter(name='yb',
               nature='internal',
               type='real',
               value='(ymb*cmath.sqrt(2))/vev',
               texname='\\text{yb}')

yc = Parameter(name='yc',
               nature='internal',
               type='real',
               value='(ymc*cmath.sqrt(2))/vev',
               texname='\\text{yc}')

ye = Parameter(name='ye',
               nature='internal',
               type='real',
               value='(yme*cmath.sqrt(2))/vev',
               texname='\\text{ye}')

ym = Parameter(name='ym',
               nature='internal',
               type='real',
               value='(ymm*cmath.sqrt(2))/vev',
               texname='\\text{ym}')

yt = Parameter(name='yt',
               nature='internal',
               type='real',
               value='(ymt*cmath.sqrt(2))/vev',
               texname='\\text{yt}')

ytau = Parameter(name='ytau',
                 nature='internal',
                 type='real',
                 value='(ymtau*cmath.sqrt(2))/vev',
                 texname='\\text{ytau}')

muH = Parameter(name='muH',
                nature='internal',
                type='real',
                value='cmath.sqrt(lam*vev**2)',
                texname='\\mu')

I1x31 = Parameter(name='I1x31',
                  nature='internal',
                  type='complex',
                  value='yb*complexconjugate(CKM1x3)',
                  texname='\\text{I1x31}')

I1x32 = Parameter(name='I1x32',
                  nature='internal',
                  type='complex',
                  value='yb*complexconjugate(CKM2x3)',
                  texname='\\text{I1x32}')

I1x33 = Parameter(name='I1x33',
                  nature='internal',
                  type='complex',
                  value='yb*complexconjugate(CKM3x3)',
                  texname='\\text{I1x33}')

I2x12 = Parameter(name='I2x12',
                  nature='internal',
                  type='complex',
                  value='yc*complexconjugate(CKM2x1)',
                  texname='\\text{I2x12}')

I2x13 = Parameter(name='I2x13',
                  nature='internal',
                  type='complex',
                  value='yt*complexconjugate(CKM3x1)',
                  texname='\\text{I2x13}')

I2x22 = Parameter(name='I2x22',
                  nature='internal',
                  type='complex',
                  value='yc*complexconjugate(CKM2x2)',
                  texname='\\text{I2x22}')

I2x23 = Parameter(name='I2x23',
                  nature='internal',
                  type='complex',
                  value='yt*complexconjugate(CKM3x2)',
                  texname='\\text{I2x23}')

I2x32 = Parameter(name='I2x32',
                  nature='internal',
                  type='complex',
                  value='yc*complexconjugate(CKM2x3)',
                  texname='\\text{I2x32}')

I2x33 = Parameter(name='I2x33',
                  nature='internal',
                  type='complex',
                  value='yt*complexconjugate(CKM3x3)',
                  texname='\\text{I2x33}')

I3x21 = Parameter(name='I3x21',
                  nature='internal',
                  type='complex',
                  value='CKM2x1*yc',
                  texname='\\text{I3x21}')

I3x22 = Parameter(name='I3x22',
                  nature='internal',
                  type='complex',
                  value='CKM2x2*yc',
                  texname='\\text{I3x22}')

I3x23 = Parameter(name='I3x23',
                  nature='internal',
                  type='complex',
                  value='CKM2x3*yc',
                  texname='\\text{I3x23}')

I3x31 = Parameter(name='I3x31',
                  nature='internal',
                  type='complex',
                  value='CKM3x1*yt',
                  texname='\\text{I3x31}')

I3x32 = Parameter(name='I3x32',
                  nature='internal',
                  type='complex',
                  value='CKM3x2*yt',
                  texname='\\text{I3x32}')

I3x33 = Parameter(name='I3x33',
                  nature='internal',
                  type='complex',
                  value='CKM3x3*yt',
                  texname='\\text{I3x33}')

I4x13 = Parameter(name='I4x13',
                  nature='internal',
                  type='complex',
                  value='CKM1x3*yb',
                  texname='\\text{I4x13}')

I4x23 = Parameter(name='I4x23',
                  nature='internal',
                  type='complex',
                  value='CKM2x3*yb',
                  texname='\\text{I4x23}')

I4x33 = Parameter(name='I4x33',
                  nature='internal',
                  type='complex',
                  value='CKM3x3*yb',
                  texname='\\text{I4x33}')



from importlib.resources import files, as_file
import logging
from ufo_model_loader.common import setup_logging, JSONLook, UFOModelLoaderError
from ufo_model_loader.commands import load_model, export_model
import argparse

# create the top-level parser
parser = argparse.ArgumentParser(prog='ufo_model_loader')

# Add options common to all subcommands
parser.add_argument('--verbosity', '-v', type=str, choices=[
                    'debug', 'info', 'critical'], default='info', help='Set verbosity level')
parser.add_argument('--debug', '-d', action='store_true', default=False,
                    help='Shortcut for --verbosity debug')
parser.add_argument('--quiet', '-q', action='store_true', default=False,
                    help='Shortcut for --verbosity critical')
parser.add_argument('--input_model', '-i', metavar='input_model_path', type=str, required=True,
                    help='UFO directory model name <UFO_directory> or or JSON file path <JSON_file_path>.json to load.')
parser.add_argument('--restriction_name', '-r', metavar='restriction_name', type=str, default=None,
                    help="""Restriction name to consider. For UFO, it will look for a file restrict_<restriction_name>.dat in the UFO model directory. If not given, it will look for restrict_default.dat and if not found, it will use no restriction. 
                            For JSON it will use the param card called <JSON_model_file_name>_<restriction_name>.json file and if not specified, then <JSON_model_file_name>_default.json, and nothing if not found.""")
parser.add_argument('--simplify', '-s', metavar='simplify', action=argparse.BooleanOptionalAction, default=True,
                    help='Simplify the model given the restriction by merging identical couplings and disabling zero contributions. Default = %(default)s')
parser.add_argument('--wrap_indices_in_lorentz_structures', metavar='wrap_indices_in_lorentz_structures', action=argparse.BooleanOptionalAction, default=False,
                    help='Wrap indices in Lorentz structures. This is in particular useful for spin-2 models, mapping <1 or 2>00<p_id> conventions to idx(1 or 2, p_id). Default = %(default)s')
parser.add_argument('--output_model_path', '-o', metavar='output_model_path', type=str, default=None,
                    help='Output JSON file path <output_model_path>.json to write the model to, with input param card written as <output_model_path>_param_card.json. If not given, it will be written to the current working directory with the name of the input model.')
parser.add_argument('--json_look', '-j', metavar='json_look', type=str, default='verbose', choices=['compact', 'pretty', 'verbose'],
                    help="JSON look to use when writing JSON files. One of 'compact', 'pretty' or 'verbose'. Default = %(default)s")
parser.add_argument('--logging_format', '-l', metavar='logging_format', type=str, default='short', choices=['none', 'min', 'short', 'long'],
                    help="Logging prefix format. One of 'none', 'min', 'short' or 'long'. Default = %(default)s")
parser.add_argument('--overwrite', '-w', action=argparse.BooleanOptionalAction, default=False,
                    help='Allow overwrite of the output files if they exist. Default = %(default)s')


def main():
    args = parser.parse_args()

    console_handler = setup_logging(args.logging_format)
    match args.verbosity.lower():
        case 'debug':
            console_handler.setLevel(logging.DEBUG)
        case 'info':
            console_handler.setLevel(logging.INFO)
        case 'critical':
            console_handler.setLevel(logging.CRITICAL)
        case _:
            raise UFOModelLoaderError(
                f"Invalid verbosity level: {args.verbosity}")

    if args.debug:
        console_handler.setLevel(logging.DEBUG)

    if args.quiet:
        console_handler.setLevel(logging.CRITICAL)

    match args.json_look.lower():
        case 'compact':
            json_look = JSONLook.COMPACT
        case 'pretty':
            json_look = JSONLook.PRETTY
        case 'verbose':
            json_look = JSONLook.VERBOSE
        case _:
            raise UFOModelLoaderError(f"Invalid JSON look: {args.json_look}")

    # Load the model
    model, input_param_card = load_model(
        input_model_path=args.input_model,
        restriction_name=args.restriction_name,
        simplify_model=args.simplify,
        wrap_indices_in_lorentz_structures=args.wrap_indices_in_lorentz_structures,
    )
    # Export the model
    export_model(
        model=model,
        input_param_card=input_param_card,
        output_model_path=args.output_model_path,
        json_look=json_look,
        allow_overwrite=args.overwrite,
    )


if __name__ == "__main__":
    main()



import re
import os
from pprint import pformat
from typing import Callable

from ufo_model_loader.common import UFOModelLoaderError, UFOModelLoaderWarning, UFOMODELLOADER_WARNINGS_ISSUED, logger  # type: ignore

SYMBOLICA_AVAILABLE = True
try:
    os.environ["SYMBOLICA_HIDE_BANNER"] = "1"
    import symbolica  # type: ignore
    from symbolica import Expression, E, S, T, AtomType # type: ignore
    _ = Expression.parse("1")  # Just to trigger the banner
    del os.environ["SYMBOLICA_HIDE_BANNER"]
except BaseException as e:
    logger.critical(
        "Symbolica could not be imported and is currently mandatory in UFOModelLoader: %s", e)
    raise UFOModelLoaderError(
        "Symbolica could not be imported and is currently mandatory in UFOModelLoader: %s", e)
    SYMBOLICA_AVAILABLE = False


def expression_to_string_safe(expr: Expression, canonical=True) -> str:
    try:
        if canonical:
            return expr.to_canonical_string()
        else:
            return expr.format(
                terms_on_new_line=False,
                color_top_level_sum=False,
                color_builtin_symbols=False,
                print_finite_field=False,
                explicit_rational_polynomial=False,
                number_thousands_separator=None,
                multiplication_operator='*',
                square_brackets_for_function=False,
                num_exp_as_superscript=False,
            )
    except Exception as exception:  # pylint: disable=broad-except
        raise UFOModelLoaderError(
            "Symbolica (@%s)failed to cast expression to string:\n%s\nwith exception:\n%s", symbolica.__file__, expr, exception)


def expression_to_string(expr: Expression | None, canonical=True) -> str | None:
    if expr is None:
        return None
    try:
        return expression_to_string_safe(expr, canonical)
    except Exception as exception:  # pylint: disable=broad-except
        logger.critical("%s", exception)
        return None


def replace_from_sqrt(expr: Expression) -> Expression:
    expr = expr.replace(Expression.parse(
        'sqrt(x__)'), Expression.parse('x__^(1/2)'), repeat=True)
    str_expr = expression_to_string(expr)
    if str_expr is None or re.match(r'\^\(\d+/\d+\)', str_expr):
        raise UFOModelLoaderError(
            "Exponentiation with real arguments not supported in model expressions: %s", str_expr)
    return expr


def replace_pseudo_floats(expression: str) -> str:

    def rationalize_float(fl: re.Match[str]) -> Expression:
        fl_eval: float = eval(fl.group())
        # Work around a bug for 0.0 in symbolica
        rationalized_fl = Expression.num(
            fl_eval, 1e-13) if fl_eval != 0. else SBE.num(0)  # type: ignore
        rationalized_fl_eval: float = eval(str(rationalized_fl)+'.')
        if UFOModelLoaderWarning.FloatInExpression not in UFOMODELLOADER_WARNINGS_ISSUED:
            UFOMODELLOADER_WARNINGS_ISSUED.add(
                UFOModelLoaderWarning.FloatInExpression)
            logger.warning(
                f"Expression contains the following floating point values in the expression: {fl.group()} (mapped to {rationalized_fl})")
            logger.warning(
                "It is typically best for the UFO model to be adjusted and declare these floats as internal constant parameters instead, or have them written as a fraction.")
        valid_cast = True
        if abs(rationalized_fl_eval-fl_eval) != 0.:
            if abs(fl_eval) != 0.:
                if abs(rationalized_fl_eval-fl_eval)/abs(fl_eval) > 1.0e-12:
                    valid_cast = False
            else:
                if abs(rationalized_fl_eval-fl_eval) > 1.0e-12:
                    valid_cast = False
        if not valid_cast:
            raise UFOModelLoaderError(
                "The float value %s in the expression could not be cast to a rational number %s = %.16e", fl.group(), rationalized_fl, rationalized_fl_eval)

        return rationalized_fl

    modified_expression = re.sub(
        r'(\d+\.\d+e[+-]?\d+)', lambda x: f'({rationalize_float(x)})', expression)
    modified_expression = re.sub(
        r'(\d+\.\d+)', lambda x: f'({rationalize_float(x)})', expression)
    modified_expression = re.sub(r'(\d+)\.', r'\1', modified_expression)

    return modified_expression


def parse_python_expression_safe(expr: str) -> Expression:

    sanitized_expr = expr.replace('**', '^')\
        .replace('cmath.sqrt', 'sqrt')\
        .replace('cmath.pi', '𝜋')\
        .replace('math.sqrt', 'sqrt')\
        .replace('math.pi', '𝜋')
    sanitized_expr = replace_pseudo_floats(sanitized_expr)
    try:
        sb_expr = Expression.parse(sanitized_expr, default_namespace='UFO')
        sb_expr = sb_expr.replace(
            E('UFO::complex(x_,y_)'), E('x_+y_ 1𝑖'), repeat=True)
    except Exception as exception:  # pylint: disable=broad-except
        raise UFOModelLoaderError(
            "Symbolica (@%s) failed to parse expression:\n%s\nwith exception:\n%s", symbolica.__file__, sanitized_expr, exception)

    # sb_expr_processed = replace_to_sqrt(sb_expr)
    sb_expr_processed = replace_from_sqrt(sb_expr)

    return sb_expr_processed


def parse_python_expression(expr: str | None) -> Expression | None:
    if expr is None:
        return None
    try:
        return parse_python_expression_safe(expr)
    except Exception as exception:  # pylint: disable=broad-except
        logger.critical("%s", exception)
        return None


def evaluate_symbolica_expression(expr: Expression, evaluation_variables: dict[Expression, complex], evaluation_functions: dict[Callable[[Expression], Expression], Callable[[list[complex]], complex]]) -> complex | None:
    try:
        return evaluate_symbolica_expression_safe(expr, evaluation_variables, evaluation_functions)
    except UFOModelLoaderError:
        return None


def evaluate_symbolica_expression_safe(expr: Expression, evaluation_variables: dict[Expression, complex], evaluation_functions: dict[Callable[[Expression], Expression], Callable[[list[complex]], complex]]) -> complex:
    try:
        res: complex = expr.evaluate_complex(  # type: ignore
            evaluation_variables, evaluation_functions)  # type: ignore
        # Small adjustment to avoid havin -0. in either the real or imaginary part
        return complex(0 if abs(res.real) == 0. else res.real,  # type: ignore
                       0 if abs(res.imag) == 0. else res.imag)  # type: ignore
    except BaseException as e:
        err_msg = "Symbolica (@%s) failed to evaluate expression:\n%s\nwith exception:\n%s.\nVariables:\n%s\nFunctions:\n%s" % (
            symbolica.__file__,
            expression_to_string(expr), e,
            pformat({expression_to_string(k): v for k,
                    v in evaluation_variables.items()}),
            pformat([expression_to_string(k()).replace('()', '')  # type: ignore
                    for k in evaluation_functions.keys()])  # type: ignore
        )
        logger.exception("%s", err_msg)
        raise UFOModelLoaderError(err_msg)

def wrap_indices(structure: Expression) -> Expression:
    
    dummy, idx, f_, w___, x_, z___ = E('UFO::dummy'), E('UFO::idx'), E('f_'), E('w___'), E('x_'), E('z___')
    
    def wrap_index(index):
        if index > 999:
            return idx(int(str(index)) // 1000,int(str(index)) % 1000)
        elif index >= 0:
            return idx(1,index)
        else:
            return dummy(abs(int(str(index))))
        
    wrapped_structure = structure.replace(
        f_(w___, x_, z___), 
        f_(w___, x_.hold(T().map(wrap_index)), z___),
        x_.req_type(AtomType.Num), level_range=(0,0), repeat=True
    )
    
    return wrapped_structure


# This file is part of the UFO.
#
# This file contains definitions for functions that
# are extensions of the cmath library, and correspond
# either to functions that are in cmath, but inconvenient
# to access from there (e.g. z.conjugate()),
# or functions that are simply not defined.
#
#
from __future__ import absolute_import
__date__ = "22 July 2010"
__author__ = "claude.duhr@durham.ac.uk"

from cmath import cos, sin, acos, asin
from .object_library import all_functions, Function

#
# shortcuts for functions from cmath
#

complexconjugate = Function(name = 'complexconjugate',
                            arguments = ('z',),
                            expression = 'z.conjugate()')


re = Function(name = 're',
              arguments = ('z',),
              expression = 'z.real')

im = Function(name = 'im',
              arguments = ('z',),
              expression = 'z.imag')


# New functions (trigonometric)

sec = Function(name = 'sec',
             arguments = ('z',),
             expression = '1./cos(z)')

asec = Function(name = 'asec',
             arguments = ('z',),
             expression = 'acos(1./z)')

csc = Function(name = 'csc',
             arguments = ('z',),
             expression = '1./sin(z)')

acsc = Function(name = 'acsc',
             arguments = ('z',),
             expression = 'asin(1./z)')




# type: ignore
from __future__ import absolute_import
from .object_library import all_parameters, Parameter

from .function_library import complexconjugate, re, im, csc, sec, acsc, asec

import os

# Specify here how many scalars need to be considered
if os.environ.get('UFO_SCALARS_MODEL_N_SCALARS') is not None:
    try:
        N_SCALARS = int(os.environ['UFO_SCALARS_MODEL_N_SCALARS'])
    except ValueError:
        raise ValueError(
            "Environment variable UFO_SCALARS_MODEL_N_SCALARS must be an integer")
    print("Loading UFO scalars model with %d scalars" % N_SCALARS)
else:
    N_SCALARS = 3

# Add all possible three-point interactions. User can filter undesired ones in vertices.py if needed.
if os.environ.get('UFO_SCALARS_MODEL_N_POINT_INTERACTIONS') is not None:
    try:
        N_POINT_INTERACTIONS = [
            int(x) for x in os.environ['UFO_SCALARS_MODEL_N_POINT_INTERACTIONS'].split(',')]
    except ValueError:
        raise ValueError(
            "Environment variable UFO_SCALARS_MODEL_N_POINT_INTERACTIONS must be a comma-separated list of integers")
    print("Loading UFO scalars model with n-point interactions, n=[%s]" %
          '|'.join(str(x) for x in N_POINT_INTERACTIONS))
else:
    N_POINT_INTERACTIONS = [3, 4, 5, 6, 7, 8, 9, 10]
# N_POINT_INTERACTIONS = [3,4]

# This is a default parameter object representing 0.
ZERO = Parameter(name='ZERO',
                 nature='internal',
                 type='real',
                 value='0.0',
                 texname='0')

# Introduce the scalar coupling lambda
lam = Parameter(name='lam',
                nature='external',
                type='real',
                value=1.0,
                texname='\\lambda',
                lhablock='SCALARS',
                lhacode=[1337])

# Assign masses and widths to the scalars
scalar_mass_parameters = []
scalar_width_parameters = []
for i_scalar in range(N_SCALARS):
    # Set the first scalar with id 0 to always be massless
    if i_scalar == 0:
        scalar_mass_parameters.append(ZERO)
        scalar_width_parameters.append(ZERO)
        continue
    # remember that the constructor automatically adds the objects to all_parameters
    scalar_mass_parameters.append(Parameter(name='mass_scalar_%s' % i_scalar,
                                            nature='external',
                                            type='real',
                                            value=1.0*i_scalar,
                                            texname='m_%d' % i_scalar,
                                            lhablock='SCALARS',
                                            lhacode=[2001+i_scalar]))
    scalar_width_parameters.append(Parameter(name='width_scalar_%s' % i_scalar,
                                             nature='external',
                                             type='real',
                                             value=0.0,
                                             texname='w_%d' % i_scalar,
                                             lhablock='SCALARS',
                                             lhacode=[3001+i_scalar]))

# User-defined parameters. Keep key SM parameters as their use may be hardcoded in some generators
aEWM1 = Parameter(name='aEWM1',
                  nature='external',
                  type='real',
                  value=132.50698,
                  texname='\\text{aEWM1}',
                  lhablock='SMINPUTS',
                  lhacode=[1])

Gf = Parameter(name='Gf',
               nature='external',
               type='real',
               value=0.0000116639,
               texname='G_f',
               lhablock='SMINPUTS',
               lhacode=[2])

aS = Parameter(name='aS',
               nature='external',
               type='real',
               value=0.118,
               texname='\\alpha _s',
               lhablock='SMINPUTS',
               lhacode=[3])

aEW = Parameter(name='aEW',
                nature='internal',
                type='real',
                value='1/aEWM1',
                texname='\\alpha _{\\text{EW}}')

G = Parameter(name='G',
              nature='internal',
              type='real',
              value='2*cmath.sqrt(aS)*cmath.sqrt(cmath.pi)',
              texname='G')

ee = Parameter(name='ee',
               nature='internal',
               type='real',
               value='2*cmath.sqrt(aEW)*cmath.sqrt(cmath.pi)',
               texname='e')



from __future__ import annotations
from symbolica import Expression, E, S  # type: ignore
from collections import Counter

import importlib
import os
from pathlib import Path
import sys
from typing import Any, Callable
from enum import StrEnum
import json
import cmath

from ufo_model_loader.common import UFOModelLoaderError, DATA_PATH, logger, UFOMODELLOADER_WARNINGS_ISSUED, UFOModelLoaderWarning, verbose_json_dump, Colour, JSONLook, optionally_lower_external_parameter_name  # type: ignore
from ufo_model_loader.symbolica_processing import SYMBOLICA_AVAILABLE, expression_to_string_safe, parse_python_expression_safe, evaluate_symbolica_expression, parse_python_expression, expression_to_string, evaluate_symbolica_expression_safe, wrap_indices  # type: ignore
from ufo_model_loader.param_card import ParamCard  # type: ignore


pjoin = os.path.join

if SYMBOLICA_AVAILABLE:
    from symbolica import Expression, S  # type: ignore


class ParameterType(StrEnum):
    COMPLEX = 'complex'
    REAL = 'real'


class ParameterNature(StrEnum):
    INTERNAL = 'internal'
    EXTERNAL = 'external'


class SerializableVertexRule(object):
    def __init__(self, name: str, particles: list[str], color_structures: list[str], lorentz_structures: list[str], couplings: list[list[str | None]]):
        self.name: str = name
        self.particles: list[str] = particles
        self.color_structures: list[str] = color_structures
        self.lorentz_structures: list[str] = lorentz_structures
        self.couplings: list[list[str | None]] = couplings

    @staticmethod
    def from_vertex_rule(vertex_rule: VertexRule) -> SerializableVertexRule:
        return SerializableVertexRule(
            vertex_rule.name,
            [particle.name for particle in vertex_rule.particles],
            [expression_to_string_safe(c)
             for c in vertex_rule.color_structures],
            [lorentz.name for lorentz in vertex_rule.lorentz_structures],
            [
                [
                    None if (i_col, j_lor) not in vertex_rule.couplings else vertex_rule.couplings[(
                        i_col, j_lor)].name
                    for j_lor in range(len(vertex_rule.lorentz_structures))
                ] for i_col in range(len(vertex_rule.color_structures))
            ]
        )

    @staticmethod
    def from_dict(dict_repr: dict[str, Any]) -> SerializableVertexRule:
        return SerializableVertexRule(
            dict_repr['name'],
            dict_repr['particles'],
            dict_repr['color_structures'],
            dict_repr['lorentz_structures'],
            dict_repr['couplings']
        )


class VertexRule(object):
    def __init__(self, name: str, particles: list[Particle], color_structures: list[str], lorentz_structures: list[LorentzStructure], couplings: dict[tuple[int, int], Coupling]):
        self.name: str = name
        self.particles: list[Particle] = particles
        self.color_structures: list[Expression] = [parse_python_expression_safe(
            color_structure) for color_structure in color_structures]
        self.lorentz_structures: list[LorentzStructure] = lorentz_structures
        self.couplings: dict[tuple[int, int], Coupling] = couplings

    @staticmethod
    def from_ufo_object(model: Model, ufo_object: Any) -> VertexRule:
        return VertexRule(
            ufo_object.name,
            [model.get_particle(particle.name)
             for particle in ufo_object.particles],
            ufo_object.color,
            [model.get_lorentz_structure(lorentz.name)
             for lorentz in ufo_object.lorentz],
            {(i, j): model.get_coupling(coupling.name)
             for (i, j), coupling in ufo_object.couplings.items()}
        )

    @staticmethod
    def from_serializable_vertex_rule(model: Model, serializable_vertex_rule: SerializableVertexRule) -> VertexRule:
        return VertexRule(
            serializable_vertex_rule.name,
            [model.get_particle(particle_name)
             for particle_name in serializable_vertex_rule.particles],
            serializable_vertex_rule.color_structures,
            [model.get_lorentz_structure(
                lorentz_name) for lorentz_name in serializable_vertex_rule.lorentz_structures],
            dict(((i, j), model.get_coupling(serializable_vertex_rule.couplings[i][j]))  # type: ignore
                 for i in range(len(serializable_vertex_rule.color_structures))
                 for j in range(len(serializable_vertex_rule.lorentz_structures)) if serializable_vertex_rule.couplings[i][j] is not None)
        )

    def to_serializable_vertex_rule(self) -> SerializableVertexRule:
        return SerializableVertexRule.from_vertex_rule(self)


class Propagator(object):
    def __init__(self, name: str, particle: Particle, numerator: str, denominator: str):
        self.name: str = name
        self.particle: Particle = particle
        self.numerator: Expression = parse_python_expression_safe(
            numerator)
        self.denominator: Expression = parse_python_expression_safe(
            denominator)

    @staticmethod
    def from_particle(particle: Particle, gauge: str) -> Propagator:

        name = f'{particle.name}_prop{gauge}'
        if particle.spin == -1:  # ghost
            numerator = 'complex(0,1)'
            if particle.is_massive():
                denominator = f'P(1)**2-{particle.mass.name}**2'
            else:
                denominator = 'P(1)**2'
        elif particle.spin == 1:  # scalar 2s+1=1
            if particle.is_massive():
                numerator = 'complex(0,1)'
                denominator = f'P(1)**2-{particle.mass.name}**2'
            else:
                numerator = 'complex(0,1)'
                denominator = 'P(1)**2'
        elif particle.spin == 2:  # spinor 2s+1=2
            if particle.is_massive():
                if particle.pdg_code < 0:
                    numerator = f'complex(0,1)*(-PSlash(1,2)+{particle.mass.name}*Identity(1,2))'
                else:
                    numerator = f'complex(0,1)*(PSlash(2,1)+{particle.mass.name}*Identity(2,1))'
                denominator = f'P(1)**2-{particle.mass.name}**2'
            else:
                if particle.pdg_code < 0:
                    numerator = '-complex(0,1)*PSlash(1,2)'
                else:
                    numerator = 'complex(0,1)*PSlash(2,1)'
                denominator = 'P(1)**2'
        elif particle.spin == 3:  # vector 2s+1=3
            if particle.is_massive():
                numerator = f'complex(0,1)*(-Metric(1,2)+P(1)*P(2)/{particle.mass.name}**2)'
                denominator = f'P(1)**2-{particle.mass.name}**2'
            else:
                denominator = 'P(1)**2'
                if gauge == 'Feynman':
                    numerator = 'complex(0,1)*(-Metric(1,2))'
                else:
                    raise UFOModelLoaderError(
                        f'Gauge {gauge} not implemented for vector particles')
        elif particle.spin == 5:  # tensor particle 2s+1=4
            if particle.is_massive():
                numerator = 'MassiveTensorPropagatorNumeratorNotImplemented'
                denominator = 'MassiveTensorPropagatorDenominatorNotImplemented'
            else:
                # Assume "dim" is a parameter of the model. Otherwise user could have defined his own propagator anyway.
                numerator = 'complex(0,1)*((1/2)*( (-2*Metric(1001, 2001)*Metric(1002, 2002))/(-2+dim) + Metric(1001,2002)*Metric(2001,1002) + Metric(1001,1002)*Metric(2001,2002) ))'
                denominator = 'P(1)**2'
        else:
            numerator = 'HigherSpinPropagatorNumeratorNotImplemented'
            denominator = 'HigherSpinPropagatorDenominatorNotImplemented'
            # raise UFOModelLoaderError(
            #    f'Particle spin {particle.spin} not implemented')

        return Propagator(
            name,
            particle,
            numerator,
            denominator
        )

    @staticmethod
    def from_serializable_propagator(model: Model, serializable_propagator: SerializablePropagator, ) -> Propagator:
        return Propagator(
            serializable_propagator.name,
            model.get_particle(serializable_propagator.particle),
            serializable_propagator.numerator,
            serializable_propagator.denominator
        )

    def to_serializable_propagator(self) -> SerializablePropagator:
        return SerializablePropagator.from_propagator(self)


class SerializablePropagator(object):
    def __init__(self, name: str, particle: str, numerator: str, denominator: str):
        self.name: str = name
        self.particle: str = particle
        self.numerator: str = numerator
        self.denominator: str = denominator

    @staticmethod
    def from_propagator(propagator: Propagator) -> SerializablePropagator:
        return SerializablePropagator(propagator.name, propagator.particle.name, expression_to_string_safe(propagator.numerator), expression_to_string_safe(propagator.denominator))

    @staticmethod
    def from_dict(dict_repr: dict[str, Any]) -> SerializablePropagator:
        return SerializablePropagator(
            dict_repr['name'],
            dict_repr['particle'],
            dict_repr['numerator'],
            dict_repr['denominator']
        )


class SerializableCoupling(object):
    def __init__(self, name: str, expression: str, orders: list[tuple[str, int]], value: tuple[float, float] | None):
        self.name: str = name
        self.expression: str = expression
        self.value: tuple[float, float] | None = value
        self.orders: list[tuple[str, int]] = orders

    @staticmethod
    def from_coupling(coupling: Coupling) -> SerializableCoupling:
        return SerializableCoupling(
            coupling.name,
            expression_to_string_safe(coupling.expression),
            list(coupling.orders.items()),
            None if coupling.value is None else (
                coupling.value.real, coupling.value.imag)
        )

    @staticmethod
    def from_dict(dict_repr: dict[str, Any]) -> SerializableCoupling:
        return SerializableCoupling(
            dict_repr['name'],
            dict_repr['expression'],
            dict_repr['orders'],
            dict_repr['value']
        )


class Coupling(object):
    def __init__(self, name: str, expression: str, orders: dict[str, int], value: complex | None = None):
        self.name: str = name
        self.expression: Expression = parse_python_expression_safe(
            expression)
        self.value: complex | None = value
        self.orders: dict[str, int] = orders

    @staticmethod
    def from_ufo_object(ufo_object: Any) -> Coupling:
        expression = ufo_object.value
        if isinstance(ufo_object.value, dict):
            if list(ufo_object.value.keys()) == [0,]:
                expression = ufo_object.value[0]
            else:
                if UFOModelLoaderWarning.DroppingEpsilonTerms not in UFOMODELLOADER_WARNINGS_ISSUED:
                    UFOMODELLOADER_WARNINGS_ISSUED.add(
                        UFOModelLoaderWarning.DroppingEpsilonTerms)
                    logger.warning(
                        "Only finite terms of order ε^0 are retained in model expressions")
                expression = ufo_object.value.get(0, '0')
        return Coupling(ufo_object.name, expression, ufo_object.order)

    @staticmethod
    def from_serializable_coupling(serializable_coupling: SerializableCoupling) -> Coupling:
        return Coupling(
            serializable_coupling.name,
            serializable_coupling.expression,
            dict(serializable_coupling.orders),
            None if serializable_coupling.value is None else complex(
                *serializable_coupling.value)
        )

    def to_serializable_coupling(self) -> SerializableCoupling:
        return SerializableCoupling.from_coupling(self)


class SerializableLorentzStructure(object):
    def __init__(self, name: str, spins: list[int], structure: str):
        self.name: str = name
        self.spins: list[int] = spins
        self.structure: str = structure

    @staticmethod
    def from_lorentz_structure(lorentz_structure: LorentzStructure) -> SerializableLorentzStructure:
        return SerializableLorentzStructure(
            lorentz_structure.name,
            lorentz_structure.spins,
            expression_to_string_safe(lorentz_structure.structure)
        )

    @staticmethod
    def from_dict(dict_repr: dict[str, Any]) -> SerializableLorentzStructure:
        return SerializableLorentzStructure(
            dict_repr['name'],
            dict_repr['spins'],
            dict_repr['structure']
        )


class LorentzStructure(object):
    def __init__(self, name: str, spins: list[int], structure: str):
        self.name: str = name
        self.spins: list[int] = spins
        self.structure: Expression = parse_python_expression_safe(
            structure)

    @staticmethod
    def from_ufo_object(ufo_object: Any) -> LorentzStructure:
        return LorentzStructure(ufo_object.name, ufo_object.spins, ufo_object.structure)

    @staticmethod
    def from_serializable_lorentz_structure(serializable_lorentz_structure: SerializableLorentzStructure) -> LorentzStructure:
        return LorentzStructure(
            serializable_lorentz_structure.name,
            serializable_lorentz_structure.spins,
            serializable_lorentz_structure.structure
        )

    def to_serializable_lorentz_structure(self) -> SerializableLorentzStructure:
        return SerializableLorentzStructure.from_lorentz_structure(self)


class SerializableParameter(object):
    def __init__(self, lhablock: str | None, lhacode: tuple[int, ...] | None, name: str, nature: str, parameter_type: str, value: tuple[float, float] | None, expression: str | None):
        self.lhablock: str | None = lhablock
        self.lhacode: tuple[int, ...] | None = lhacode
        self.name: str = name
        self.nature: str = nature
        self.parameter_type: str = parameter_type
        self.value: tuple[float, float] | None = value
        self.expression: str | None = expression

    @staticmethod
    def from_parameter(parameter: Parameter) -> SerializableParameter:
        return SerializableParameter(
            parameter.lhablock, tuple(
                parameter.lhacode) if parameter.lhacode is not None else None, parameter.name,
            str(parameter.nature), str(
                parameter.parameter_type),
            None if parameter.value is None else (
                parameter.value.real, parameter.value.imag),
            expression_to_string(parameter.expression)
        )

    @staticmethod
    def from_dict(dict_repr: dict[str, Any]) -> SerializableParameter:
        return SerializableParameter(
            dict_repr['lhablock'], tuple(
                dict_repr['lhacode']) if dict_repr['lhacode'] is not None else None, dict_repr['name'],
            dict_repr['nature'], dict_repr['parameter_type'], dict_repr['value'],
            dict_repr['expression']
        )


class Parameter(object):

    # Useful for assigning typed defaults
    @staticmethod
    def default() -> Parameter:
        return Parameter(lhablock='DUMMY', lhacode=(0,), name='DUMMY', nature=ParameterNature.INTERNAL, parameter_type=ParameterType.REAL, value=0., expression='0')

    def __init__(self, lhablock: str | None, lhacode: tuple[int, ...] | None, name: str, nature: ParameterNature, parameter_type: ParameterType, value: complex | None, expression: str | None):
        self.lhablock: str | None = lhablock
        self.lhacode: tuple[int, ...] | None = lhacode
        self.name: str = name
        self.nature: ParameterNature = nature
        self.parameter_type: ParameterType = parameter_type
        self.value: complex | None = value
        self.expression: Expression | None = parse_python_expression(
            expression)

    @staticmethod
    def from_ufo_object(ufo_object: Any) -> Parameter:
        # This ugly hack is to fix an annoying persistent typo in many NLO models
        if ufo_object.nature == 'interal':
            ufo_object.nature = 'internal'

        if ufo_object.nature == 'external':
            param_value = complex(ufo_object.value)
            param_expression: str | None = None
        elif ufo_object.nature == 'internal':
            expression: Any = ufo_object.value
            if isinstance(ufo_object.value, dict):
                if list(ufo_object.value.keys()) == [0,]:
                    expression = ufo_object.value[0]
                else:
                    if UFOModelLoaderWarning.DroppingEpsilonTerms not in UFOMODELLOADER_WARNINGS_ISSUED:
                        UFOMODELLOADER_WARNINGS_ISSUED.add(
                            UFOModelLoaderWarning.DroppingEpsilonTerms)
                        logger.warning(
                            "Only finite terms of order ε^0 are retained in model expressions")
                    expression = ufo_object.value.get(0, '0')
            param_expression = expression
            try:
                param_value = complex(expression)
                param_expression = None
            except ValueError:
                param_value = None
        else:
            raise UFOModelLoaderError(
                f"Invalid parameter nature '{ufo_object.nature}'")

        return Parameter(
            ufo_object.lhablock if hasattr(ufo_object, 'lhablock') else None,
            (None if ufo_object.lhacode is None else tuple(ufo_object.lhacode)
             ) if hasattr(ufo_object, 'lhacode') else None,
            ufo_object.name,
            ParameterNature(ufo_object.nature), ParameterType(ufo_object.type),
            param_value, param_expression
        )

    @staticmethod
    def from_serializable_parameter(serializable_parameter: SerializableParameter) -> Parameter:
        return Parameter(
            serializable_parameter.lhablock, serializable_parameter.lhacode, serializable_parameter.name,
            ParameterNature(serializable_parameter.nature), ParameterType(
                serializable_parameter.parameter_type),
            None if serializable_parameter.value is None else complex(
                *serializable_parameter.value),
            serializable_parameter.expression
        )

    def to_serializable_parameter(self) -> SerializableParameter:
        return SerializableParameter.from_parameter(self)

    def __str__(self) -> str:
        if self.nature == ParameterNature.INTERNAL:
            return f"Internal parameter({self.name}, {self.expression})"
        else:
            return f"External parameter({self.name}, {self.value})"


class SerializableParticle(object):
    def __init__(self, pdg_code: int, name: str, antiname: str, spin: int, color: int, mass: str, width: str, texname: str, antitexname: str, charge: float, ghost_number: int, lepton_number: int, y_charge: int):
        self.pdg_code: int = pdg_code
        self.name: str = name
        self.antiname: str = antiname
        self.spin: int = spin
        self.color: int = color
        self.mass: str = mass
        self.width: str = width
        self.texname: str = texname
        self.antitexname: str = antitexname
        self.charge: float = charge
        self.ghost_number: int = ghost_number
        self.lepton_number: int = lepton_number
        self.y_charge: int = y_charge

    @classmethod
    def from_particle(cls, particle: Particle) -> SerializableParticle:
        return SerializableParticle(
            particle.pdg_code, particle.name, particle.antiname, particle.spin, particle.color,
            particle.mass.name,
            particle.width.name,
            particle.texname, particle.antitexname,
            particle.charge,
            particle.ghost_number,
            particle.lepton_number,
            particle.y_charge
        )

    @classmethod
    def from_dict(cls, dict_repr: dict[str, Any]) -> SerializableParticle:
        return SerializableParticle(
            dict_repr['pdg_code'], dict_repr['name'], dict_repr['antiname'], dict_repr['spin'], dict_repr['color'],
            dict_repr['mass'],
            dict_repr['width'],
            dict_repr['texname'], dict_repr['antitexname'],
            dict_repr['charge'],
            dict_repr['ghost_number'],
            dict_repr['lepton_number'],
            dict_repr['y_charge']
        )


class Particle(object):
    def __init__(self, pdg_code: int, name: str, antiname: str, spin: int, color: int, mass: Parameter, width: Parameter, texname: str, antitexname: str, charge: float, ghost_number: int, lepton_number: int, y_charge: int):
        self.pdg_code: int = pdg_code
        self.name: str = name
        self.antiname: str = antiname
        self.spin: int = spin
        self.color: int = color
        self.mass: Parameter = mass
        self.width: Parameter = width
        self.texname: str = texname
        self.antitexname: str = antitexname
        self.charge: float = charge
        self.ghost_number: int = ghost_number
        self.lepton_number: int = lepton_number
        self.y_charge: int = y_charge

    @staticmethod
    def default() -> Particle:
        return Particle(0, '', '', 1, 1, Parameter.default(), Parameter.default(), '', '', 0, 0, 0, 0)

    def is_ghost(self) -> bool:
        return self.ghost_number != 0

    def is_massive(self) -> bool:
        return self.mass.value is None or abs(self.mass.value) > 0.

    @staticmethod
    def sanitize_texname(texname: str) -> str:
        return texname
        # If the rendered does not support this, uncomment below        
        #return texname.replace('~', 'x').replace('_', '')

    @staticmethod
    def from_ufo_object(model: Model, ufo_object: Any) -> Particle:

        return Particle(
            ufo_object.pdg_code, ufo_object.name, ufo_object.antiname, ufo_object.spin, ufo_object.color,
            model.get_parameter(ufo_object.mass.name),
            model.get_parameter(ufo_object.width.name),
            Particle.sanitize_texname(ufo_object.texname),
            Particle.sanitize_texname(ufo_object.antitexname),
            ufo_object.charge,
            ufo_object.GhostNumber,
            ufo_object.LeptonNumber,
            ufo_object.Y if hasattr(ufo_object, 'Y') else 0
        )

    @staticmethod
    def from_serializable_particle(model: Model, serializable_particle: SerializableParticle) -> Particle:
        return Particle(
            serializable_particle.pdg_code, serializable_particle.name, serializable_particle.antiname, serializable_particle.spin, serializable_particle.color,
            model.get_parameter(serializable_particle.mass),
            model.get_parameter(serializable_particle.width),
            serializable_particle.texname, serializable_particle.antitexname,
            serializable_particle.charge,
            serializable_particle.ghost_number,
            serializable_particle.lepton_number,
            serializable_particle.y_charge
        )

    def to_serializable_particle(self) -> SerializableParticle:
        return SerializableParticle.from_particle(self)

    def get_pdg_code(self) -> int:
        return self.pdg_code

    def get_anti_pdg_code(self, model: Model) -> int:
        return self.get_anti_particle(model).get_pdg_code()

    def get_anti_particle(self, model: Model) -> Particle:
        return model.get_particle(self.antiname)


class Order(object):
    def __init__(self, expansion_order: int, hierarchy: int, name: str):
        self.expansion_order: int = expansion_order
        self.hierarchy: int = hierarchy
        self.name: str = name

    @staticmethod
    def from_ufo_object(ufo_object: Any) -> Order:
        return Order(ufo_object.expansion_order, ufo_object.hierarchy, ufo_object.name)

    @staticmethod
    def from_dict(dict_repr: Any) -> Order:
        return Order(dict_repr['expansion_order'], dict_repr['hierarchy'], dict_repr['name'])


class SerializableModel(object):

    def __init__(self, name: str):
        self.name: str = name
        self.restriction: str | None = None
        self.orders: list[Order] = []
        self.parameters: list[SerializableParameter] = []
        self.particles: list[SerializableParticle] = []
        self.propagators: list[SerializablePropagator] = []
        self.lorentz_structures: list[SerializableLorentzStructure] = []
        self.couplings: list[SerializableCoupling] = []
        self.vertex_rules: list[SerializableVertexRule] = []

    @staticmethod
    def from_model(model: Model) -> SerializableModel:
        serializable_model = SerializableModel(model.name)
        serializable_model.restriction = model.restriction
        serializable_model.orders = model.orders
        serializable_model.parameters = [
            parameter.to_serializable_parameter() for parameter in model.parameters]
        serializable_model.particles = [
            particle.to_serializable_particle() for particle in model.particles]
        serializable_model.propagators = [
            propagator.to_serializable_propagator() for propagator in model.propagators]
        serializable_model.lorentz_structures = [lorentz_structure.to_serializable_lorentz_structure(
        ) for lorentz_structure in model.lorentz_structures]
        serializable_model.couplings = [
            coupling.to_serializable_coupling() for coupling in model.couplings]
        serializable_model.vertex_rules = [
            vertex_rule.to_serializable_vertex_rule() for vertex_rule in model.vertex_rules]
        return serializable_model

    def to_dict(self) -> dict:
        return {
            'name': self.name,
            'restriction': self.restriction,
            'orders': [order.__dict__ for order in self.orders],
            'parameters': [parameter.__dict__ for parameter in self.parameters],
            'particles': [particle.__dict__ for particle in self.particles],
            'propagators': [propagator.__dict__ for propagator in self.propagators],
            'lorentz_structures': [lorentz_structure.__dict__ for lorentz_structure in self.lorentz_structures],
            'couplings': [coupling.__dict__ for coupling in self.couplings],
            'vertex_rules': [vertex_rule.__dict__ for vertex_rule in self.vertex_rules]
        }

    def to_json(self, json_look: JSONLook = JSONLook.VERBOSE) -> str:
        return verbose_json_dump(self.to_dict(), json_look=json_look)

    @staticmethod
    def from_json(json_str: str) -> SerializableModel:
        json_model = json.loads(json_str)
        serializable_model = SerializableModel(json_model['name'])
        serializable_model.restriction = json_model['restriction']
        serializable_model.orders = [Order.from_dict(
            order) for order in json_model['orders']]
        serializable_model.parameters = [SerializableParameter.from_dict(
            parameter) for parameter in json_model['parameters']]
        serializable_model.particles = [SerializableParticle.from_dict(
            particle) for particle in json_model['particles']]
        serializable_model.propagators = [SerializablePropagator.from_dict(
            propagator) for propagator in json_model['propagators']]
        serializable_model.lorentz_structures = [SerializableLorentzStructure.from_dict(
            lorentz_structure) for lorentz_structure in json_model['lorentz_structures']]
        serializable_model.couplings = [SerializableCoupling.from_dict(
            coupling) for coupling in json_model['couplings']]
        serializable_model.vertex_rules = [SerializableVertexRule.from_dict(
            vertex_rule) for vertex_rule in json_model['vertex_rules']]
        return serializable_model


class Model(object):

    # Because parameters are now sorted accorded to their dependency,
    # it should not be necessary to go deeper than 1 level of recursion
    MAX_ALLOWED_RECURSION_IN_PARAMETER_EVALUATION = 1

    def __init__(self, name: str):
        self.name: str = name
        self.restriction: str | None = None
        self.orders: list[Order] = []
        self.parameters: list[Parameter] = []
        self.particles: list[Particle] = []
        self.propagators: list[Propagator] = []
        self.lorentz_structures: list[LorentzStructure] = []
        self.couplings: list[Coupling] = []
        self.vertex_rules: list[VertexRule] = []

        self.name_to_position: dict[str, dict[str | int, int]] = {}

    def get_full_name(self) -> str:
        return f"{self.name}-{'full' if self.restriction is None else self.restriction}"

    # We must delay the building of model functions to make sure Symbolica has been registered at this point.
    @staticmethod
    def model_function_reglog(xs: list[complex]) -> complex:
        arg: list[float] = [xs[0].real, xs[0].imag]
        if abs(arg[0]) == 0.:
            arg[0] = 0.
        if abs(arg[1]) == 0.:
            arg[1] = 0.
        if arg == [0., 0.]:
            return 0.
        else:
            return cmath.log(complex(arg[0], arg[1]))

    @staticmethod
    def model_function_reglogp(xs: list[complex]) -> complex:
        arg: list[float] = [xs[0].real, xs[0].imag]
        if abs(arg[0]) == 0.:
            arg[0] = 0.
        if abs(arg[1]) == 0.:
            arg[1] = 0.
        if arg == [0., 0.]:
            return 0.
        else:
            if arg[0] < 0. and arg[1] < 0.:
                return cmath.log(complex(arg[0], arg[1])) + complex(0., 2*cmath.pi)
            else:
                return cmath.log(complex(arg[0], arg[1]))

    @staticmethod
    def model_function_reglogm(xs: list[complex]) -> complex:
        arg: list[float] = [xs[0].real, xs[0].imag]
        if abs(arg[0]) == 0.:
            arg[0] = 0.
        if abs(arg[1]) == 0.:
            arg[1] = 0.
        if arg == [0., 0.]:
            return 0.
        else:
            if arg[0] < 0. and arg[1] > 0.:
                return cmath.log(complex(arg[0], arg[1])) - complex(0., 2*cmath.pi)
            else:
                return cmath.log(complex(arg[0], arg[1]))

    @classmethod
    def set_model_functions(cls):
        cls.model_functions: dict[Callable[[Expression], Expression], Callable[[list[complex]], complex]] = {  # type: ignore
            S('UFO::Theta'): lambda xs: 0. if xs[0].real < 0. else 1.,
            S('UFO::tan'): lambda xs: cmath.tan(xs[0]),
            S('UFO::acos'): lambda xs: cmath.acos(xs[0]),
            S('UFO::asin'): lambda xs: cmath.asin(xs[0]),
            S('UFO::atan'): lambda xs: cmath.atan(xs[0]),
            S('UFO::complexconjugate'): lambda xs: xs[0].conjugate(),
            S('UFO::complex'): lambda xs: xs[0]+complex(0, 1)*xs[1],
            # TUFO::hese functions are typically only used in NLO models
            S('UFO::cond'): lambda xs: xs[1] if abs(xs[0]) == 0. else xs[2],
            S('UFO::reglog'): lambda xs: Model.model_function_reglog(xs),
            S('UFO::reglogp'): lambda xs: Model.model_function_reglogp(xs),
            S('UFO::reglogm'): lambda xs: Model.model_function_reglogm(xs),
        }

    @classmethod
    def get_model_functions(cls) -> dict[Callable[[Expression], Expression], Callable[[list[complex]], complex]]:
        if not hasattr(cls, 'model_functions'):
            cls.set_model_functions()
        return dict(cls.model_functions)  # type: ignore

    @classmethod
    def set_model_variables(cls):
        cls.model_variables: dict[Expression, complex] = {  # type: ignore
            S('UFO::I'): complex(0, 1),
            S('UFO::pi'): cmath.pi
        }

    @classmethod
    def get_model_variables(cls) -> dict[Expression, complex]:
        if not hasattr(cls, 'model_variables'):
            cls.set_model_variables()
        return dict(cls.model_variables)  # type: ignore

    def is_empty(self) -> bool:
        return self.name == 'NotLoaded' or len(self.particles) == 0

    def get_internal_parameters(self) -> list[Parameter]:
        return [parameter for parameter in self.parameters if parameter.nature == ParameterNature.INTERNAL]

    def get_external_parameters(self) -> list[Parameter]:
        return [parameter for parameter in self.parameters if parameter.nature == ParameterNature.EXTERNAL]

    def get_coupling_orders(self) -> list[str]:
        return list(set(sum([list(coupling.orders.keys()) for coupling in self.couplings], [])))

    def sort_parameters(self):
        """ Sort parameters to linearize their evaluations, meaning all dependent parameters of parameter i appear before i """

        sorted_parameters: list[Parameter] = self.get_external_parameters()

        remaining_parameters: list[Parameter] = self.get_internal_parameters()
        for p in list(remaining_parameters):
            if p.expression is None:
                if p.value is None:
                    raise UFOModelLoaderError(
                        "Parameter {} has no expression and no value".format(p.name))
                else:
                    sorted_parameters.append(
                        remaining_parameters.pop(remaining_parameters.index(p)))

        param_variables: dict[str, dict[str, Any]] = {
            p.name: {
                'var': S(f'UFO::{p.name}'),
                'dependent_params': p.expression.get_all_symbols(False)
            } for p in remaining_parameters if p.expression is not None
        }

        while len(remaining_parameters) > 0:
            n_added: int = 0
            for p in list(remaining_parameters):
                if not any(
                        any(dep_param == param_variables[p.name]['var']
                            for p in remaining_parameters)
                        for dep_param in param_variables[p.name]['dependent_params']):
                    n_added += 1
                    sorted_parameters.append(
                        remaining_parameters.pop(remaining_parameters.index(p)))
            if n_added == 0:
                raise UFOModelLoaderError(
                    "Circular dependency in parameters of model {}", self.name)

        self.parameters = sorted_parameters
        self.sync_name_to_position_dict()

    @staticmethod
    def from_ufo_model(ufo_model_path: str) -> Model:

        model_path = os.path.abspath(
            pjoin(DATA_PATH, 'models', ufo_model_path))
        if not os.path.isdir(model_path):
            model_path = os.path.abspath(ufo_model_path)
        if not os.path.isdir(model_path):
            raise UFOModelLoaderError(
                f"UFO model '{ufo_model_path}' cannot be found")
        sys.path.insert(0, os.path.dirname(model_path))
        sys.path.insert(0, model_path)
        model_name = os.path.basename(model_path)
        ufo_model = importlib.import_module(model_name)
        del sys.path[0]
        del sys.path[0]

        # Note that this approach does not work because of relative imports in the UFO model __init__.py file
        # spec = importlib.util.spec_from_file_location(os.path.basename(model_path), pjoin(model_path,'__init__.py'))
        # ufo_model = importlib.util.module_from_spec(spec)
        # spec.loader.exec_module(ufo_model)

        model = Model(model_name)

        # Load coupling orders
        model.orders = [Order.from_ufo_object(
            order) for order in ufo_model.all_orders]
        model.name_to_position['orders'] = {
            order.name: i for i, order in enumerate(model.orders)}
        # Load parameters
        model.parameters = [Parameter.from_ufo_object(
            param) for param in ufo_model.all_parameters]
        model.name_to_position['parameters'] = {
            param.name: i for i, param in enumerate(model.parameters)}
        if hasattr(ufo_model, 'all_CTparameters'):
            ct_parameters = [Parameter.from_ufo_object(
                param) for param in ufo_model.all_CTparameters]
            model.name_to_position['parameters'].update({
                param.name: len(model.parameters)+i for i, param in enumerate(ct_parameters)})
            model.parameters.extend(ct_parameters)

        # Load particles
        model.particles = [Particle.from_ufo_object(
            model, particle) for particle in ufo_model.all_particles]
        model.name_to_position['particles'] = {
            particle.name: i for i, particle in enumerate(model.particles)}
        model.name_to_position['particles_from_PDG'] = {
            particle.pdg_code: i for i, particle in enumerate(model.particles)}
        model.set_case_insensitive_particle_dictionary()

        # Load propagators
        model.propagators = [Propagator.from_particle(
            particle, 'Feynman') for particle in model.particles]
        model.name_to_position['propagators'] = {
            propagator.name: i for i, propagator in enumerate(model.propagators)}
        # Load Lorentz structures
        model.lorentz_structures = [LorentzStructure.from_ufo_object(
            lorentz) for lorentz in ufo_model.all_lorentz]
        model.name_to_position['lorentz_structures'] = {
            lorentz.name: i for i, lorentz in enumerate(model.lorentz_structures)}
        # Load Couplings
        model.couplings = [Coupling.from_ufo_object(
            coupling) for coupling in ufo_model.all_couplings]
        model.name_to_position['couplings'] = {
            coupling.name: i for i, coupling in enumerate(model.couplings)}
        # Load Vertices (ignore counterterms)
        model.vertex_rules = [VertexRule.from_ufo_object(
            model, vertex) for vertex in ufo_model.all_vertices if vertex.__class__.__name__ != 'CTVertex']
        model.name_to_position['vertex_rules'] = {
            vertex_rule.name: i for i, vertex_rule in enumerate(model.vertex_rules)}

        model.sort_parameters()

        return model

    def sync_name_to_position_dict(self):
        self.name_to_position.clear()
        self.name_to_position['orders'] = {
            order.name: i for i, order in enumerate(self.orders)}
        self.name_to_position['parameters'] = {
            param.name: i for i, param in enumerate(self.parameters)}
        self.name_to_position['particles'] = {
            particle.name: i for i, particle in enumerate(self.particles)}
        self.name_to_position['particles_from_PDG'] = {
            particle.pdg_code: i for i, particle in enumerate(self.particles)}
        self.set_case_insensitive_particle_dictionary()

        self.name_to_position['propagators'] = {
            propagator.name: i for i, propagator in enumerate(self.propagators)}
        self.name_to_position['lorentz_structures'] = {
            lorentz.name: i for i, lorentz in enumerate(self.lorentz_structures)}
        self.name_to_position['couplings'] = {
            coupling.name: i for i, coupling in enumerate(self.couplings)}
        self.name_to_position['vertex_rules'] = {
            vertex_rule.name: i for i, vertex_rule in enumerate(self.vertex_rules)}

    def set_case_insensitive_particle_dictionary(self):
        case_sensitivity_check: dict[str, str] = {}
        for k in self.name_to_position['particles'].keys():
            if str(k).lower() in case_sensitivity_check:
                raise UFOModelLoaderError(
                    f"Particle {k} is case-insensitive equivalent to {case_sensitivity_check[str(k).lower()]}")
            case_sensitivity_check[str(k).lower()] = str(k)

        self.name_to_position['particles_case_insensitive'] = {
            particle.name.lower(): i for i, particle in enumerate(self.particles)}

    @staticmethod
    def from_serializable_model(serializable_model: SerializableModel) -> Model:
        model = Model(serializable_model.name)
        model.restriction = serializable_model.restriction
        model.orders = serializable_model.orders

        model.name_to_position['orders'] = {
            order.name: i for i, order in enumerate(model.orders)}
        model.parameters = [Parameter.from_serializable_parameter(
            param) for param in serializable_model.parameters]
        model.name_to_position['parameters'] = {
            param.name: i for i, param in enumerate(model.parameters)}
        model.particles = [Particle.from_serializable_particle(
            model, particle) for particle in serializable_model.particles]
        model.name_to_position['particles'] = {
            particle.name: i for i, particle in enumerate(model.particles)}
        model.name_to_position['particles_from_PDG'] = {
            particle.pdg_code: i for i, particle in enumerate(model.particles)}
        model.set_case_insensitive_particle_dictionary()

        model.propagators = [Propagator.from_serializable_propagator(
            model, propagator) for propagator in serializable_model.propagators]
        model.lorentz_structures = [LorentzStructure.from_serializable_lorentz_structure(
            lorentz) for lorentz in serializable_model.lorentz_structures]
        model.name_to_position['lorentz_structures'] = {
            lorentz.name: i for i, lorentz in enumerate(model.lorentz_structures)}
        model.couplings = [Coupling.from_serializable_coupling(
            coupling) for coupling in serializable_model.couplings]
        model.name_to_position['couplings'] = {
            coupling.name: i for i, coupling in enumerate(model.couplings)}
        model.vertex_rules = [VertexRule.from_serializable_vertex_rule(
            model, vertex_rule) for vertex_rule in serializable_model.vertex_rules]
        model.name_to_position['vertex_rules'] = {
            vertex_rule.name: i for i, vertex_rule in enumerate(model.vertex_rules)}

        return model

    def to_serializable_model(self) -> SerializableModel:
        return SerializableModel.from_model(self)

    @staticmethod
    def from_json(json_str: str) -> Model:
        return Model.from_serializable_model(SerializableModel.from_json(json_str))

    def to_json(self, json_look: JSONLook = JSONLook.VERBOSE) -> str:
        return self.to_serializable_model().to_json(json_look=json_look)

    def get_vertices_from_particles(self, particles: list[Particle]) -> list[VertexRule]:
        return [vertex_rule for vertex_rule in self.vertex_rules if Counter(vertex_rule.particles) == Counter(particles)]

    def get_order(self, order_name: str) -> Order:
        return self.orders[self.name_to_position['orders'][order_name]]

    def get_parameter(self, parameter_name: str) -> Parameter:
        return self.parameters[self.name_to_position['parameters'][parameter_name]]

    def get_parameter_from_lha_specification(self, lhablock: str, lhacode: tuple[int, ...]) -> Parameter | None:
        for parameter in self.parameters:
            if parameter.lhablock is None:
                continue
            if parameter.lhablock.lower() == lhablock.lower() and parameter.lhacode == lhacode:
                return parameter
        return None

    def get_particle(self, particle_name: str) -> Particle:
        return self.particles[self.name_to_position['particles_case_insensitive'][particle_name.lower()]]

    def get_particle_from_pdg(self, pdg: int) -> Particle:
        return self.particles[self.name_to_position['particles_from_PDG'][pdg]]

    def get_propagator(self, propagator_name: str) -> Propagator:
        return self.propagators[self.name_to_position['propagators'][propagator_name]]

    def get_lorentz_structure(self, lorentz_name: str) -> LorentzStructure:
        return self.lorentz_structures[self.name_to_position['lorentz_structures'][lorentz_name]]

    def get_coupling(self, coupling_name: str) -> Coupling:
        return self.couplings[self.name_to_position['couplings'][coupling_name]]

    def get_vertex_rule(self, vertex_name: str) -> VertexRule:
        return self.vertex_rules[self.name_to_position['vertex_rules'][vertex_name]]

    def remove_zero_couplings(self) -> tuple[list[Coupling], list[VertexRule]]:

        removed_couplings: list[Coupling] = []
        remaining_couplings: list[Coupling] = []
        for coupling in self.couplings:
            if coupling.value == 0.:
                removed_couplings.append(coupling)
            else:
                remaining_couplings.append(coupling)
        self.couplings = remaining_couplings

        removed_vertex_rules: list[VertexRule] = []
        remaining_vertex_rules: list[VertexRule] = []
        for vertex_rule in self.vertex_rules:
            vertex_rule.couplings = {
                key: coupling
                for key, coupling in vertex_rule.couplings.items() if coupling.value != 0.
            }
            if len(vertex_rule.couplings) == 0:
                removed_vertex_rules.append(vertex_rule)
            else:
                remaining_vertex_rules.append(vertex_rule)
        self.vertex_rules = remaining_vertex_rules
        self.sync_name_to_position_dict()

        return (removed_couplings, removed_vertex_rules)

    def get_parameter_values(self) -> dict[Expression, complex]:
        parameter_map: dict[Expression, complex] = {}
        for parameter in self.parameters:
            if parameter.value is None:
                raise UFOModelLoaderError(
                    f"The value of parameter {parameter.name} has not been set yet")
            parameter_map[S(f'UFO::{parameter.name}')] = parameter.value
        return parameter_map

    def update_coupling_values(self) -> None:
        parameter_values = self.get_model_variables()
        parameter_values.update(self.get_parameter_values())
        for coupling in self.couplings:
            coupling.value = evaluate_symbolica_expression_safe(
                coupling.expression, parameter_values, self.get_model_functions())

    def update_internal_parameters(self, input_card: InputParamCard) -> None:
        evaluation_variables: dict[Expression,
                                   complex] = self.get_model_variables()
        for parameter in self.get_external_parameters():
            parameter.value = input_card[optionally_lower_external_parameter_name(
                parameter.name)]
            evaluation_variables[S(f'UFO::{parameter.name}')] = parameter.value

        # Collect constant internal parameters such as ZERO
        for parameter in self.get_internal_parameters():
            if parameter.expression is None:
                if parameter.value is None:
                    raise UFOModelLoaderError(
                        "Internal parameter '{parameter.name}' has no value nor expression.")
                evaluation_variables[S(f'UFO::{parameter.name}')] = parameter.value

        # Now update all other dependent variables
        # Note that this is done in a loop because some parameters may depend on other parameters
        evaluation_round: int = 0
        while True:
            evaluation_round += 1
            found_new_evaluation = False
            found_unevaluated = False
            for parameter in self.get_internal_parameters():
                parameter_Expression = S(f'UFO::{parameter.name}')
                if parameter_Expression not in evaluation_variables:
                    if parameter.expression is None:
                        raise UFOModelLoaderError(
                            f"Internal parameter '{parameter.name}' has no value nor expression.")
                    # print([expression_to_string_safe(f(S('UFO::x')))
                    #       for f in self.get_model_functions()])
                    # print(parameter.expression)
                    if evaluation_round == Model.MAX_ALLOWED_RECURSION_IN_PARAMETER_EVALUATION:
                        eval_result: complex | None = evaluate_symbolica_expression_safe(
                            parameter.expression, evaluation_variables, self.get_model_functions())
                    else:
                        eval_result = evaluate_symbolica_expression(
                            parameter.expression, evaluation_variables, self.get_model_functions())
                    if eval_result is not None:
                        parameter.value = eval_result
                        evaluation_variables[parameter_Expression] = parameter.value
                        found_new_evaluation = True
                    else:
                        found_unevaluated = True

            if not found_unevaluated:
                break
            if not found_new_evaluation:
                logger.critical(
                    "Could not evaluate all internal parameters. Reprocessing evaluation now to show latest error:")
                for parameter in self.get_internal_parameters():
                    if parameter.expression is not None:
                        eval_result = evaluate_symbolica_expression_safe(
                            parameter.expression, evaluation_variables, self.get_model_functions())
                raise UFOModelLoaderError(
                    "Could not evaluate all internal parameters.")

            if evaluation_round >= Model.MAX_ALLOWED_RECURSION_IN_PARAMETER_EVALUATION:
                raise UFOModelLoaderError(
                    "Maximum number of allowed recursions in parameter evaluation reached.")

    def wrap_indices_in_lorentz_structures(self) -> None:
        for lorentz_structure in self.lorentz_structures:
            lorentz_structure.structure = wrap_indices(lorentz_structure.structure)
        for propagator in self.propagators:
            propagator.numerator = wrap_indices(propagator.numerator)
            propagator.denominator = wrap_indices(propagator.denominator)

    def apply_input_param_card(self, input_card: InputParamCard, simplify: bool = False, update: bool = True) -> None:
        # from pprint import pprint
        # pprint(input_card)
        external_parameters = self.get_external_parameters()
        zero_parameters: list[str] = []
        for param in external_parameters:
            if optionally_lower_external_parameter_name(param.name) not in input_card:
                if param.value is not None:
                    logger.debug(
                        "Parameter '%s' not set in input parameter card. It will be added now with its default value of %s", param.name, param.value)
                    input_card[optionally_lower_external_parameter_name(
                        param.name)] = param.value
                else:
                    raise UFOModelLoaderError(
                        f"Parameter '{param.name}' is external with no default value, but not set in input parameter card.")
            else:
                param.value = input_card[optionally_lower_external_parameter_name(
                    param.name)]

            if param.value == 0. and simplify:
                zero_parameters.append(param.name)
                param.nature = ParameterNature.INTERNAL
                param.expression = parse_python_expression_safe('ZERO')
                del input_card[optionally_lower_external_parameter_name(
                    param.name)]

        if len(zero_parameters) > 0:
            logger.info("The following %s%d external parameters%s were forced to zero by the restriction card:\n%s%s%s", Colour.GREEN, len(zero_parameters), Colour.END,
                        Colour.BLUE, ', '.join(zero_parameters), Colour.END)

        if update:
            self.update_internal_parameters(input_card)

            self.update_coupling_values()

            if simplify:
                removed_couplings, removed_vertices = self.remove_zero_couplings()
                if len(removed_couplings) > 0:
                    logger.info("A total of %s%d couplings%s have been %sremoved%s due restriction card specification", Colour.GREEN, len(
                        removed_couplings), Colour.END, Colour.GREEN, Colour.END)
                if len(removed_vertices) > 0:
                    logger.info("A total of %s%d vertices%s have been %sremoved%s due restriction card specification:\n%s%s%s",
                                Colour.GREEN, len(
                                    removed_vertices), Colour.END, Colour.GREEN, Colour.END, Colour.BLUE,
                                ', '.join(f"{v_r.name} -> ({'|'.join(p.name for p in v_r.particles)})" for v_r in removed_vertices), Colour.END)


class InputParamCard(dict[str, complex]):

    @staticmethod
    def from_param_card(param_card: ParamCard, model: Model | None = None) -> InputParamCard:
        input_param_card = InputParamCard()
        all_parameters, _restrictions = param_card.analyze_param_card(model)
        for param_name, locations in all_parameters.items():
            if len(locations) != 1:
                raise UFOModelLoaderError(
                    f"Composite parameter defined across multiple LHA entries not supported. Parameter name: {param_name} and locations: {locations}")
            block_name, block_lhaid = locations[0]
            param_value = param_card.get_value(
                block_name, block_lhaid, 0.0)
            if not isinstance(param_value, float):
                raise UFOModelLoaderError(
                    f"Parameter value must be float, this is not the case for parameter '{param_name}' with value '{param_value}'.")
            input_param_card[param_name] = complex(param_value, 0.)
        return input_param_card

    @staticmethod
    def from_model(model: Model) -> InputParamCard:
        input_param_card = InputParamCard()
        for param in model.get_external_parameters():
            if param.value is not None:
                input_param_card[optionally_lower_external_parameter_name(
                    param.name)] = param.value
            else:
                raise UFOModelLoaderError(
                    f"Model '{model.name}' cannot be loaded without any restriction card since at least one parameter, '{param.name}' lacks a default value.")
        return input_param_card

    @staticmethod
    def from_json(json_str: str) -> InputParamCard:
        res = InputParamCard()
        json_dict = json.loads(json_str)
        for param_name, param_value in json_dict.items():
            res[param_name] = complex(param_value[0], param_value[1])
        return res

    @staticmethod
    def from_json_file(json_path: str) -> InputParamCard:
        with open(json_path, 'r') as f:
            return InputParamCard.from_json(f.read())

    def to_json(self, json_look: JSONLook = JSONLook.VERBOSE) -> str:
        return verbose_json_dump({k: [v.real, v.imag] for k, v in self.items()}, json_look=json_look)
